import warnings
import tensorflow as tf 
import keras
import typing 
from tensorflow.python.framework import composite_tensor

from molcraft import ops


class GraphTensorBatchEncoder(tf.experimental.ExtensionTypeBatchEncoder):
    
    def batch(self, spec: 'GraphTensor.Spec', batch_size: int | None):
        """Batches spec.
        """

        def batch_field(f):
            if isinstance(f, tf.TensorSpec):
                return tf.TensorSpec(
                    shape=[None] + f.shape[1:],
                    dtype=f.dtype
                )
            elif isinstance(f, tf.RaggedTensorSpec):
                return tf.RaggedTensorSpec(
                    shape=[batch_size, None] + f.shape[1:],
                    dtype=f.dtype,
                    ragged_rank=1,
                    row_splits_dtype=f.row_splits_dtype
                )
            elif isinstance(f, tf.TypeSpec):
                return f.__batch_encoder__.batch(f, batch_size)
            return f
        fields = dict(spec.__dict__)
        # Pop context fields as they will be batched differently
        context_fields = fields.pop('context')
        batched_fields = tf.nest.map_structure(batch_field, fields)
        batched_spec = object.__new__(type(spec))
        batched_context_fields = tf.nest.map_structure(
            lambda spec: tf.TensorSpec([batch_size] + spec.shape, spec.dtype), 
            context_fields
        )
        batched_spec.__dict__.update({'context': batched_context_fields})
        batched_spec.__dict__.update(batched_fields)
        return batched_spec
    
    def unbatch(self, spec: 'GraphTensor.Spec'):
        """Unbatches spec.
        """

        def unbatch_field(f):   
            if isinstance(f, tf.TensorSpec):
                return tf.TensorSpec(
                    shape=[None] + f.shape[1:],
                    dtype=f.dtype
                )
            elif isinstance(f, tf.RaggedTensorSpec):
                return tf.RaggedTensorSpec(
                    shape=[None] + f.shape[2:],
                    dtype=f.dtype,
                    ragged_rank=0,
                    row_splits_dtype=f.row_splits_dtype
                )
            elif isinstance(f, tf.TypeSpec):
                return f.__batch_encoder__.unbatch(f)
            return f
        fields = dict(spec.__dict__)
        # Pop context fields as they will be unbatched differently
        context_fields = fields.pop('context')
        unbatched_fields = tf.nest.map_structure(unbatch_field, fields)
        unbatched_context_fields = tf.nest.map_structure(
            lambda spec: tf.TensorSpec(spec.shape[1:], spec.dtype), 
            context_fields
        )
        unbatched_spec = object.__new__(type(spec))
        unbatched_spec.__dict__.update({'context': unbatched_context_fields})
        unbatched_spec.__dict__.update(unbatched_fields)
        return unbatched_spec
        
    def encode(self, spec: 'GraphTensor.Spec', value: 'GraphTensor', minimum_rank: int = 0):
        """Encodes value.
        """
        unflatten = False if (is_ragged(spec) or is_scalar(spec)) else True 
        if unflatten:
            value = value.unflatten()
        value_components = tuple(value.__dict__[key] for key in spec.__dict__)
        value_components = tuple(
            x for x in tf.nest.flatten(value_components) 
            if isinstance(x, (tf.Tensor, composite_tensor.CompositeTensor))
        )
        return value_components
    
    def encoding_specs(self, spec: 'GraphTensor.Spec'):
        """Matches spec and encoded value of `encode(spec, value)`.
        """
        def encode_fields(f):
            if isinstance(f, tf.TensorSpec):
                scalar = is_scalar(spec)
                return tf.RaggedTensorSpec(
                    shape=([None] if scalar else [None, None]) + f.shape[1:], 
                    dtype=f.dtype, 
                    ragged_rank=(0 if scalar else 1),
                    row_splits_dtype=spec.context['size'].dtype
                )
            return f
        fields = dict(spec.__dict__)
        context_fields = fields.pop('context')
        encoded_fields = tf.nest.map_structure(encode_fields, fields)
        encoded_fields = {**{'context': context_fields}, **encoded_fields}
        spec_components = tuple(encoded_fields.values())
        spec_components = tuple(
            x for x in tf.nest.flatten(spec_components)
            if isinstance(x, tf.TypeSpec)
        )
        return spec_components
    
    def decode(self, spec, encoded_value):
        """Decodes encoded value.
        """
        spec_tuple = tuple(spec.__dict__.values())
        encoded_value = iter(encoded_value)
        value_tuple = [
            next(encoded_value) if isinstance(x, tf.TypeSpec) else x
            for x in tf.nest.flatten(spec_tuple)
        ]
        value_tuple = tf.nest.pack_sequence_as(spec_tuple, value_tuple)
        fields = dict(zip(spec.__dict__.keys(), value_tuple))
        value = object.__new__(spec.value_type)
        value.__dict__.update(fields)
        flatten = is_ragged(value) and not is_ragged(spec)
        if flatten:
            value = value.flatten()
        return value
    

class GraphTensor(tf.experimental.BatchableExtensionType):
    context: typing.Mapping[str, tf.Tensor]
    node: typing.Mapping[str, typing.Union[tf.Tensor, tf.RaggedTensor]]
    edge: typing.Mapping[str, typing.Union[tf.Tensor, tf.RaggedTensor]]

    __batch_encoder__ = GraphTensorBatchEncoder()

    __name__ = 'GraphTensor'

    def __validate__(self):
        if tf.executing_eagerly():
            assert 'size' in self.context, "graph.context['size'] is required."
            assert self.context['size'].dtype == tf.int32, (
                "dtype of graph.context['size'] needs to be int32."
            )
            assert 'feature' in self.node, "graph.node['feature'] is required."
            assert 'source' in self.edge, "graph.edge['source'] is required."
            assert 'target' in self.edge, "graph.edge['target'] is required."
            assert self.edge['source'].dtype == tf.int32, (
                "dtype of graph.edge['source'] needs to be int32."
            )
            assert self.edge['target'].dtype == tf.int32, (
                "dtype of graph.edge['target'] needs to be int32."
            )
            if isinstance(self.node['feature'], tf.Tensor):
                num_nodes = keras.ops.shape(self.node['feature'])[0]
            else:
                num_nodes = keras.ops.sum(self.node['feature'].row_lengths())
            assert keras.ops.sum(self.context['size']) == num_nodes, (
                "graph.node['feature'] tensor is incompatible with graph.context['size']"
            )
        
    @property 
    def spec(self):
        def unspecify_spec(s):
            if isinstance(s, tf.TensorSpec):
                return tf.TensorSpec([None] + s.shape[1:], s.dtype)
            return s
        orig_spec = tf.type_spec_from_value(self)
        fields = dict(orig_spec.__dict__)
        context_fields = fields.pop('context')
        new_spec_components = tf.nest.map_structure(unspecify_spec, fields)
        new_spec_components['context'] = context_fields
        return orig_spec.__class__(**new_spec_components)
    
    @property 
    def shape(self):
        if is_ragged(self):
            return self.node['feature'].shape 
        return self.context['size'].shape + [None] + self.node['feature'].shape[1:] 
    
    @property 
    def dtype(self):
        return self.node['feature'].dtype

    @property
    def graph_indicator(self):
        dtype = self.context['size'].dtype
        if is_scalar(self):
            return tf.zeros(tf.shape(self.node['feature'])[:1], dtype=dtype)
        num_graphs = keras.ops.shape(self.context['size'])[0]
        if is_ragged(self):
            return keras.ops.arange(num_graphs, dtype=dtype)
        return keras.ops.repeat(keras.ops.arange(num_graphs, dtype=dtype), self.context['size'])

    @property
    def num_subgraphs(self) -> tf.Tensor:
        dtype = self.context['size'].dtype
        if is_scalar(self):
            num_subgraphs = tf.constant(1, dtype=dtype)
        else:
            num_subgraphs = tf.shape(self.context['size'], out_type=dtype)[0]
        if tf.executing_eagerly():
            return int(num_subgraphs)
        return num_subgraphs
    
    @property 
    def num_nodes(self):
        num_nodes = keras.ops.shape(self.node['feature'])[0]
        if tf.executing_eagerly():
            return int(num_nodes)
        return num_nodes 
    
    @property 
    def num_edges(self):
        num_edges = keras.ops.shape(self.edge['source'])[0]
        if tf.executing_eagerly():
            return int(num_edges)
        return num_edges 
    
    def gather(self, node_attr: str, edge_type: str) -> tf.Tensor:
        if edge_type != 'source' and edge_type != 'target':
            raise ValueError
        return ops.gather(self.node[node_attr], self.edge[edge_type])
    
    def aggregate(self, edge_attr: str, edge_type: str = 'target', mode: str = 'sum') -> tf.Tensor:
        if edge_type != 'source' and edge_type != 'target':
            raise ValueError('`edge_attr` needs to be `source` or `target`.')
        edge_attr = self.edge[edge_attr]
        if 'weight' in self.edge:
            edge_attr = ops.edge_weight(edge_attr, self.edge['weight'])
        return ops.aggregate(edge_attr, self.edge[edge_type], self.num_nodes, mode=mode)
    
    def propagate(self, add_edge_feature: bool = False):
        updated_feature = ops.propagate(
            node_feature=self.node['feature'],
            edge_source=self.edge['source'],
            edge_target=self.edge['target'],
            edge_feature=self.edge.get('feature') if add_edge_feature else None,
            edge_weight=self.edge.get('weight'),
        )
        return self.update({'node': {'feature': updated_feature}})

    def flatten(self):
        if not is_ragged(self):
            raise ValueError(
                f"{self.__class__.__qualname__} instance is already flat.")
        def flatten_fn(x):
            if isinstance(x, tf.RaggedTensor):
                return x.flat_values
            return x
        edge_increment = ops.gather(
            self.node['feature'].row_starts(), self.edge['source'].value_rowids())
        edge_increment = tf.cast(
            edge_increment, dtype=self.edge['source'].dtype)
        data = to_dict(self)
        flat_values = tf.nest.map_structure(flatten_fn, data)
        flat_values['edge']['source'] += edge_increment
        flat_values['edge']['target'] += edge_increment 
        return from_dict(flat_values)
        
    def unflatten(self, *, force: bool = False):
        if is_scalar(self):
            raise ValueError(
                f"{self.__class__.__qualname__} instance is a scalar " 
                "and cannot be unflattened.")
        if is_ragged(self):
            raise ValueError(
                f"{self.__class__.__qualname__} instance is already unflat.")
        def unflatten_fn(x, value_rowids, nrows) -> tf.RaggedTensor:
            if isinstance(x, tf.Tensor):
                return tf.RaggedTensor.from_value_rowids(x, value_rowids, nrows)
            return x
        graph_indicator_node = self.graph_indicator
        graph_indicator_edge = ops.gather(graph_indicator_node, self.edge['source'])
        if force:
            sorted_indices = keras.ops.argsort(graph_indicator_edge)
        num_subgraphs = self.num_subgraphs
        unflat_values = {}
        data = to_dict(self)
        for key in tf.type_spec_from_value(self).__dict__:
            value = data[key]
            if key == 'context':
                unflat_values[key] = value
            elif key == 'node':
                unflat_values[key] = tf.nest.map_structure(
                    lambda x: unflatten_fn(x, graph_indicator_node, num_subgraphs),
                    value)
                row_starts = unflat_values[key]['feature'].row_starts()
                edge_decrement = ops.gather(row_starts, graph_indicator_edge)
                if force:
                    edge_decrement = ops.gather(edge_decrement, sorted_indices)
                    graph_indicator_edge = ops.gather(graph_indicator_edge, sorted_indices)
                edge_decrement = tf.cast(edge_decrement, dtype=self.edge['source'].dtype)
            elif key == 'edge':
                if force:
                    value = tf.nest.map_structure(lambda x: ops.gather(x, sorted_indices), value)
                value['source'] -= edge_decrement
                value['target'] -= edge_decrement
                unflat_values[key] = tf.nest.map_structure(
                    lambda x: unflatten_fn(x, graph_indicator_edge, num_subgraphs),
                    value)
        return from_dict(unflat_values)

    def update(self, values):
        data = to_dict(self)
        for outer_field, mapping in values.items():
            if outer_field == 'context':
                reference_value = data[outer_field]['size']
            elif outer_field == 'edge':
                reference_value = data[outer_field]['source']
            else:
                reference_value = data[outer_field]['feature']
            for inner_field, value in mapping.items():
                if value is None:
                    data[outer_field].pop(inner_field, None)
                    continue
                data[outer_field][inner_field] = _maybe_convert_new_value(
                    value, reference_value
                )
        return self.__class__(**data)

    def replace(self, values):
        data = to_dict(self)
        for outer_field, mapping in values.items():
            if outer_field == 'context':
                reference_value = data[outer_field]['size']
            elif outer_field == 'edge':
                reference_value = data[outer_field]['source']
            else:
                reference_value = data[outer_field]['feature']
            for inner_field, value in mapping.items():
                values[outer_field][inner_field] = _maybe_convert_new_value(
                    value, reference_value
                )
        return self.__class__(**values)
    
    def with_context(self, *args, **kwargs):
        context = (
            args[0] if len(args) and isinstance(args[0], dict) else kwargs
        )
        return self.update({'context': context})

    def __getitem__(self, index):
        if index is None and is_scalar(self):
            return self.__class__(
                context={key: value[None] for (key, value) in self.context.items()},
                node=self.node,
                edge=self.edge,
            )
        if not is_ragged(self):
            is_flat = True
            tensor = self.unflatten()
        else:
            tensor = self
            is_flat = False
        data = to_dict(tensor)
        if isinstance(index, (slice, int)):
            data = tf.nest.map_structure(lambda x: x[index], data)
        else:
            data = tf.nest.map_structure(lambda x: ops.gather(x, index), data)
        tensor = from_dict(data)
        if is_flat and not is_scalar(tensor):
            return tensor.flatten()
        return tensor
    
    def __repr__(self):
        return _repr(self)
    
    def numpy(self):
        """For now added to work with `keras.Model.predict`"""
        return self 

    class Spec:

        def __init__(
            self, 
            context: typing.Mapping[str, tf.TensorSpec | tf.RaggedTensorSpec], 
            node: typing.Mapping[str, tf.TensorSpec | tf.RaggedTensorSpec],  
            edge: typing.Mapping[str, tf.TensorSpec | tf.RaggedTensorSpec],  
        ) -> None:
            self.context = context 
            self.node = node
            self.edge = edge 

        @property
        def shape(self):
            if is_ragged(self):
                return self.node['feature'].shape 
            return self.context['size'].shape + [None] + self.node['feature'].shape[1:] 

        @classmethod
        def from_input_shape_dict(cls, input_shape: dict[str, tf.TensorShape]) -> 'GraphTensor.Spec':
            for key, value in input_shape.items():
                input_shape[key] = {k: tf.TensorShape(v) for k, v in value.items()}
            return cls(**tf.nest.map_structure(lambda s: tf.TensorSpec(s, dtype=tf.variant), input_shape)) 

        def __repr__(self):
            return _repr(self)
    
    
@tf.experimental.dispatch_for_api(tf.concat, {'values': typing.List[GraphTensor]})
def graph_tensor_concat(
    values: typing.List[GraphTensor],
    axis: int = 0,
    name: str = 'concat'
) -> GraphTensor:
    ragged = [is_ragged(v) for v in values]
    if 0 < sum(ragged) < len(ragged):
        raise ValueError(
            'Nested data of the GraphTensor instances do not have consistent '
            'types: found both tf.RaggedTensor values and tf.Tensor values.')
    else:
        ragged = ragged[0]

    if ragged:
        values = [v.flatten() for v in values]

    flat_values = [tf.nest.flatten(v, expand_composites=True) for v in values]
    flat_values = [tf.concat(f, axis=0) for f in list(zip(*flat_values))]
    num_edges = [keras.ops.shape(v.edge['source'])[0] for v in values]
    num_nodes = [keras.ops.shape(v.node['feature'])[0] for v in values]
    incr = tf.concat([[0], tf.cumsum(num_nodes)[:-1]], axis=0)
    incr = tf.repeat(incr, num_edges)
    value = tf.nest.pack_sequence_as(values[0], flat_values, expand_composites=True)

    edge_update = {
        'source': value.edge['source'] + incr,
        'target': value.edge['target'] + incr,
    }
    value = value.update(
        {
            'edge': edge_update
        },
    )
    if not ragged:
        return value
    return value.unflatten()

# TODO: Clean this up.
@tf.experimental.dispatch_for_api(tf.stack, {'values': typing.List[GraphTensor]})
def graph_tensor_stack(
    values: typing.List[GraphTensor],
    axis: int = 0,
    name: str = 'stack'
) -> GraphTensor:
    ragged = [is_ragged(v) for v in values]
    if not is_scalar(values[0]):
        raise ValueError(
            'tf.stack on a list of `GraphTensor`s is currently '
            'only supported for scalar `GraphTensor`s. '
        )
    if any(ragged):
        raise ValueError(
            'tf.stack on a list of `GraphTensor`s is currently '
            'only supported for flattened `GraphTensor`s. '
        )
    
    def concat_or_stack(k, v):
        if k.startswith('context'):
            return tf.stack(v, axis=0)
        return tf.concat(v, axis=0)

    fields = tuple(tf.type_spec_from_value(values[0]).__dict__)
    num_inner_fields = tuple(len(values[0].__dict__[field]) for field in fields)
    outer_keys = []
    for (f, num_fields) in zip(fields, num_inner_fields):
        outer_keys.extend([f] * num_fields)

    flat_values = [tf.nest.flatten(v, expand_composites=True) for v in values]
    flat_values = [concat_or_stack(k, f) for k, f in zip(outer_keys, list(zip(*flat_values)))]
    value = tf.nest.pack_sequence_as(values[0], flat_values, expand_composites=True)

    num_edges = [keras.ops.shape(v.edge['source'])[0] for v in values]
    num_nodes = [keras.ops.shape(v.node['feature'])[0] for v in values]
    incr = tf.concat([[0], tf.cumsum(num_nodes)[:-1]], axis=0)
    incr = tf.repeat(incr, num_edges)
    edge_update = {
        'source': value.edge['source'] + incr,
        'target': value.edge['target'] + incr,
    }
    value = value.update(
        {
            'edge': edge_update
        },
    )
    return value

def is_scalar(value_or_spec: GraphTensor | GraphTensor.Spec) -> bool:
    return value_or_spec.context['size'].shape.rank == 0

def is_ragged(value_or_spec: GraphTensor | GraphTensor.Spec) -> bool:
    is_ragged = isinstance(
        value_or_spec.node['feature'], (tf.RaggedTensor, tf.RaggedTensorSpec))
    if isinstance(value_or_spec, tf.RaggedTensorSpec):
        is_ragged = (
            is_ragged and value_or_spec.node['feature'].ragged_rank == 1)
    return is_ragged

def to_dict(tensor: GraphTensor) -> dict:
    spec = tf.type_spec_from_value(tensor)
    return {key: dict(tensor.__dict__[key]) for key in spec.__dict__}

def from_dict(data: dict) -> GraphTensor:
    data['context']['size'] = tf.cast(data['context']['size'], tf.int32)
    data['edge']['source'] = tf.cast(data['edge']['source'], tf.int32)
    data['edge']['target'] = tf.cast(data['edge']['target'], tf.int32)
    return GraphTensor(**data)

def is_graph(data):
    if isinstance(data, GraphTensor):
        return True 
    elif isinstance(data, dict) and 'size' in data.get('context', {}):
        return True 
    return False
    
def _maybe_convert_new_value(
    new_value: tf.Tensor | tf.RaggedTensor, 
    old_value: tf.Tensor | tf.RaggedTensor | None,
) -> tf.Tensor | tf.RaggedTensor:
    if old_value is None:
        return new_value
    is_old_ragged = isinstance(old_value, tf.RaggedTensor)
    is_new_ragged = isinstance(new_value, tf.RaggedTensor)
    if is_old_ragged and not is_new_ragged:
        new_value = old_value.with_flat_values(new_value)
    elif not is_old_ragged and is_new_ragged:
        new_value = new_value.flat_values
    return new_value

def _repr(x: GraphTensor | GraphTensor.Spec):
    if isinstance(x, GraphTensor):
        def _trepr(v: tf.Tensor | tf.RaggedTensor):
            if isinstance(v, tf.Tensor):
                return f'<tf.Tensor: shape={v.shape.as_list()}, dtype={v.dtype.name}>'
            return (
                f'<tf.RaggedTensor: shape={v.shape.as_list()}, ' 
                f'dtype={v.dtype.name}, ragged_rank={v.ragged_rank}>'
            )
    else:
        def _trepr(v: tf.TensorSpec | tf.RaggedTensorSpec):
            if isinstance(v, tf.TensorSpec):
                return f'<tf.TensorSpec: shape={v.shape.as_list()}, dtype={v.dtype.name}>'
            return (
                f'<tf.RaggedTensorSpec: shape={v.shape.as_list()}, ' 
                f'dtype={v.dtype.name}, ragged_rank={v.ragged_rank}>'
            )
        
    context_fields = f',\n        '.join([f'{k!r}: {_trepr(v)}' for k, v in x.context.items()])
    node_fields = f',\n        '.join([f'{k!r}: {_trepr(v)}'for k, v in x.node.items()])
    edge_fields = f',\n        '.join([f'{k!r}: {_trepr(v)}' for k, v in x.edge.items()])

    context_field = 'context={\n        ' + context_fields + '\n    }' 
    node_field = 'node={\n        ' + node_fields + '\n    }' 
    edge_field = 'edge={\n        ' + edge_fields + '\n    }' 

    fields = ',\n    '.join([context_field, node_field, edge_field])
    return x.__class__.__name__ + '(\n    ' + fields + '\n)'


import warnings
import collections
import numpy as np

from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem import Lipinski
from rdkit.Chem import rdDistGeom
from rdkit.Chem import rdDepictor
from rdkit.Chem import rdMolAlign
from rdkit.Chem import rdMolTransforms
from rdkit.Chem import rdPartialCharges
from rdkit.Chem import rdMolDescriptors
from rdkit.Chem import rdForceFieldHelpers
from rdkit.Chem import rdFingerprintGenerator


RDKitMol = Chem.Mol


class Mol(RDKitMol):
 
    @classmethod
    def from_encoding(cls, encoding: str, explicit_hs: bool = False, **kwargs) -> 'Mol':
        rdkit_mol = get_mol(encoding, **kwargs)
        if explicit_hs:
            rdkit_mol = Chem.AddHs(rdkit_mol) 
        rdkit_mol.__class__ = cls 
        setattr(rdkit_mol, '_encoding', encoding)
        return rdkit_mol

    @classmethod
    def cast(cls, obj: RDKitMol) -> 'Mol':
        obj.__class__ = cls 
        return obj

    @property
    def canonical_smiles(self) -> str:
        return Chem.MolToSmiles(self, canonical=True)
    
    @property 
    def encoding(self):
        return getattr(self, '_encoding', None)
        
    @property
    def bonds(self) -> list['Bond']:
        return get_bonds(self)
    
    @property
    def atoms(self) -> list['Atom']:
        return get_atoms(self)
    
    @property
    def num_conformers(self) -> int:
        return int(self.GetNumConformers())
    
    @property
    def num_atoms(self) -> int:
        return int(self.GetNumAtoms())
    
    @property 
    def num_bonds(self) -> int:
        return int(self.GetNumBonds())
    
    def get_atom(
        self, 
        atom: int | Chem.Atom
    ) -> 'Atom':
        if isinstance(atom, Chem.Atom):
            atom = atom.GetIdx()
        return Atom.cast(self.GetAtomWithIdx(int(atom)))
    
    def get_shortest_path_between_atoms(
        self, 
        atom_i: int | Chem.Atom, 
        atom_j: int | Chem.Atom
    ) -> tuple[int]:
        if isinstance(atom_i, Chem.Atom):
            atom_i = atom_i.GetIdx()
        if isinstance(atom_j, Chem.Atom):
            atom_j = atom_j.GetIdx()
        return Chem.rdmolops.GetShortestPath(
            self, int(atom_i), int(atom_j)
        )

    def get_bond_between_atoms(
        self, 
        atom_i: int | Chem.Atom, 
        atom_j: int | Chem.Atom,
    ) -> 'Bond':
        if isinstance(atom_i, Chem.Atom):
            atom_i = atom_i.GetIdx()
        if isinstance(atom_j, Chem.Atom):
            atom_j = atom_j.GetIdx()
        return Bond.cast(self.GetBondBetweenAtoms(int(atom_i), int(atom_j)))
        
    def adjacency(
        self, 
        fill: str = 'upper', 
        sparse: bool = True, 
        self_loops: bool = False, 
        dtype: str= 'int32', 
    ) -> np.ndarray | tuple[np.ndarray, np.ndarray]:
        return get_adjacency_matrix(
            self, fill=fill, sparse=sparse, self_loops=self_loops, dtype=dtype
        )

    def get_conformer(self, index: int = 0) -> 'Conformer':
        if self.num_conformers == 0:
            warnings.warn(f'{self} has no conformer. Returning None.')
            return None
        return Conformer.cast(self.GetConformer(index))
    
    def get_conformers(self) -> list['Conformer']:
        if self.num_conformers == 0:
            warnings.warn(f'{self} has no conformers. Returning an empty list.')
            return []
        return [Conformer.cast(x) for x in self.GetConformers()]
     
    def __len__(self) -> int:
        return int(self.GetNumAtoms())
    
    def _repr_png_(self) -> None:
        return None
    
    def __repr__(self) -> str:
        encoding = self.encoding or self.canonical_smiles
        return f'<{self.__class__.__name__} {encoding} at {hex(id(self))}>'


class Conformer(Chem.Conformer):

    @classmethod
    def cast(cls, obj: Chem.Conformer) -> 'Conformer':
        obj.__class__ = cls 
        return obj 

    @property 
    def index(self) -> int:
        return self.GetId()
    
    @property 
    def coordinates(self) -> np.ndarray:
        return self.GetPositions()
    
    @property
    def distances(self) -> np.ndarray:
        return Chem.rdmolops.Get3DDistanceMatrix(self.GetOwningMol())
    
    @property 
    def centroid(self) -> np.ndarray:
        return np.asarray(rdMolTransforms.ComputeCentroid(self))
    
    def adjacency(
        self, 
        fill: str = 'full',
        radius: float = None,
        sparse: bool = True, 
        self_loops: bool = False, 
        dtype: str = 'int32'
    ) -> np.ndarray | tuple[np.ndarray, np.ndarray]:
        radius = radius or np.inf
        distances = self.distances
        if not self_loops:
            np.fill_diagonal(distances, np.inf)
        within_radius = distances < radius
        if fill == 'lower':
            within_radius = np.tril(within_radius, k=-1)
        elif fill == 'upper':
            within_radius = np.triu(within_radius, k=1)
        if sparse:
            edge_source, edge_target = np.where(within_radius)
            return edge_source.astype(dtype), edge_target.astype(dtype)
        return within_radius.astype(dtype)
        

class Atom(Chem.Atom):

    @classmethod
    def cast(cls, obj: Chem.Atom) -> 'Atom':
        obj.__class__ = cls 
        return obj 

    @property 
    def index(self) -> int:
        return int(self.GetIdx())
    
    @property
    def neighbors(self) -> list['Atom']:
        return [Atom.cast(neighbor) for neighbor in self.GetNeighbors()]

    @property
    def symbol(self) -> str:
        return self.GetSymbol()

    @property
    def label(self):
        if self.HasProp('molAtomMapNumber'):
            return int(self.GetProp('molAtomMapNumber'))
        return None

    @label.setter
    def label(self, value: int) -> None:
        self.SetProp('molAtomMapNumber', str(value))
    
    def __repr__(self) -> str:
        return f'<Atom {self.GetSymbol()} at {hex(id(self))}>'


class Bond(Chem.Bond):

    @classmethod
    def cast(cls, obj: Chem.Bond) -> 'Bond':
        obj.__class__ = cls 
        return obj 
    
    @property 
    def index(self) -> int:
        return int(self.GetIdx())
    
    def __repr__(self) -> str:
        return f'<Bond {self.GetBondType().name} at {hex(id(self))}>'
    

def get_mol(
    encoding: str,
    strict: bool = True,
    assign_stereo_chemistry: bool = True,
) -> RDKitMol:
    if not isinstance(encoding, str):
        raise ValueError(
            f'Input ({encoding}) is not a SMILES or InChI string.'
        )
    if encoding.startswith('InChI'):
        mol = Chem.MolFromInchi(encoding, sanitize=False)
    else:
        mol = Chem.MolFromSmiles(encoding, sanitize=False)
    if mol is not None:
        mol = sanitize_mol(mol, strict, assign_stereo_chemistry)
    if mol is not None:
        return mol
    raise ValueError(f'Could not obtain `chem.Mol` from {encoding}.')

def get_adjacency_matrix(
    mol: RDKitMol,
    fill: str = 'full',
    sparse: bool = False,
    self_loops: bool = False,
    dtype: str = "int32",
) -> tuple[np.ndarray, np.ndarray]:
    adjacency: np.ndarray = Chem.GetAdjacencyMatrix(mol)
    if fill == 'lower':
        adjacency = np.tril(adjacency, k=-1)
    elif fill == 'upper':
        adjacency = np.triu(adjacency, k=1)
    if self_loops:
        adjacency += np.eye(adjacency.shape[0], dtype=adjacency.dtype)
    if not sparse:
        return adjacency.astype(dtype)
    edge_source, edge_target = np.where(adjacency)
    return edge_source.astype(dtype), edge_target.astype(dtype)

def sanitize_mol(
    mol: RDKitMol,
    strict: bool = True,
    assign_stereo_chemistry: bool = True,
) -> Mol:
    mol = Mol(mol)
    flag = Chem.SanitizeMol(mol, catchErrors=True)
    if flag != Chem.SanitizeFlags.SANITIZE_NONE:
        if strict:
            raise ValueError(f'Could not sanitize {mol}.')
        warnings.warn(
            f'Could not sanitize {mol}. Proceeding with partial sanitization.'
        )
        # Sanitize mol, excluding the steps causing the error previously
        Chem.SanitizeMol(mol, sanitizeOps=Chem.SanitizeFlags.SANITIZE_ALL^flag)
    if assign_stereo_chemistry:
        Chem.AssignStereochemistry(
            mol, cleanIt=True, force=True, flagPossibleStereoCenters=True)
    return mol

def get_atoms(mol: Mol) -> list[Atom]:
    return [
        Atom.cast(mol.GetAtomWithIdx(i)) 
        for i in range(mol.GetNumAtoms())
    ]

def get_bonds(mol: Mol) -> list[Bond]:
    return [
        Bond.cast(mol.GetBondWithIdx(int(i)))
        for i in range(mol.GetNumBonds())
    ]

def add_hs(mol: Mol) -> Mol:
    rdkit_mol = Chem.AddHs(mol)
    rdkit_mol.__class__ = mol.__class__
    return rdkit_mol

def remove_hs(mol: Mol) -> Mol:
    rdkit_mol = Chem.RemoveHs(mol)
    rdkit_mol.__class__ = mol.__class__
    return rdkit_mol

def get_distances(
    mol: Mol,
    fill: str = 'full',
    use_bond_order: bool = False, 
    use_atom_weights: bool = False
) -> np.ndarray:
    dist_matrix = Chem.rdmolops.GetDistanceMatrix(
        mol, useBO=use_bond_order, useAtomWts=use_atom_weights
    )
    # For disconnected nodes, a value of 1e8 is assigned to dist_matrix
    # Here we convert this large value to -1.
    # TODO: Add argument for filling disconnected node pairs.
    dist_matrix = np.where(
        dist_matrix >= 1e6, -1, dist_matrix
    )
    if fill == 'lower':
        return np.tril(dist_matrix, k=-1)
    elif fill == 'upper':
        return np.triu(dist_matrix, k=1)
    return dist_matrix

def get_shortest_paths( 
    mol: Mol, 
    radius: int, 
    self_loops: bool = False,
) -> list[list[int]]:
    paths = []
    for atom in mol.atoms:
        queue = collections.deque([(atom, [atom.index])])
        visited = set([atom.index])
        while queue:
            current_atom, path = queue.popleft()
            if len(path) > (radius + 1):
                continue
            if len(path) > 1 or self_loops:
                paths.append(path)
            for neighbor in current_atom.neighbors:
                if neighbor.index in visited:
                    continue
                visited.add(neighbor.index)
                queue.append((neighbor, path + [neighbor.index]))
    return paths

def get_periodic_table():
    return Chem.GetPeriodicTable()

def partial_charges(mol: 'Mol') -> list[float]:
    rdPartialCharges.ComputeGasteigerCharges(mol)
    return [atom.GetDoubleProp("_GasteigerCharge") for atom in mol.atoms]

def logp_contributions(mol: 'Mol') -> list[float]:
    return [i[0] for i in rdMolDescriptors._CalcCrippenContribs(mol)]

def molar_refractivity_contributions(mol: 'Mol') -> list[float]:
    return [i[1] for i in rdMolDescriptors._CalcCrippenContribs(mol)]

def total_polar_surface_area_contributions(mol: 'Mol') -> list[float]:
    return list(rdMolDescriptors._CalcTPSAContribs(mol))

def accessible_surface_area_contributions(mol: 'Mol') -> list[float]:
    return list(rdMolDescriptors._CalcLabuteASAContribs(mol)[0])

def hydrogen_acceptors(mol: 'Mol') -> list[bool]:
    h_acceptors = [i[0] for i in Lipinski._HAcceptors(mol)]
    return [atom.index in h_acceptors for atom in mol.atoms]

def hydrogen_donors(mol: 'Mol') -> list[bool]:
    h_donors = [i[0] for i in Lipinski._HDonors(mol)]
    return [atom.index in h_donors for atom in mol.atoms]

def hetero_atoms(mol: 'Mol') -> list[bool]:
    hetero_atoms = [i[0] for i in Lipinski._Heteroatoms(mol)]
    return [atom.index in hetero_atoms for atom in mol.atoms]

def rotatable_bonds(mol: 'Mol') -> list[bool]:
    rotatable_bonds = [set(x) for x in Lipinski._RotatableBonds(mol)]
    def is_rotatable(bond):
        atom_indices = {bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()}
        return atom_indices in rotatable_bonds
    return [is_rotatable(bond) for bond in mol.bonds]

def conformer_deviations(mol: Mol, fill: str = 'full') -> np.array:
    """Root mean squared deviation (RMSD) matrix"""
    num_confs = mol.num_conformers
    deviations = rdMolAlign.GetAllConformerBestRMS(mol)
    matrix = np.zeros((num_confs, num_confs))
    k = 0
    for i in range(num_confs):
        for j in range(i+1, num_confs):
            deviation = deviations[k]
            if fill == 'upper':
                matrix[i, j] = deviation 
            elif fill == 'lower':
                matrix[j, i] = deviation 
            else:
                matrix[i, j] = deviation
                matrix[j, i] = deviation
            k += 1
    return matrix

def conformer_energies(
    mol: Mol,
    method: str = 'UFF', 
) -> list[float]:
    if method == 'UFF':
        energies = _calc_uff_energies(mol)
    else:
        if method == 'MMFF':
            method += '94'
        variant = method
        energies = _calc_mmff_energies(mol, variant)
    return energies

def embed_conformers(
    mol: Mol, 
    num_conformers: int, 
    method: str = 'ETKDGv3',
    timeout: int | None = None, 
    random_seed: int | None = None, 
    **kwargs
) -> Mol:
    available_embedding_methods = {
        'ETDG': rdDistGeom.ETDG(),
        'ETKDG': rdDistGeom.ETKDG(),
        'ETKDGv2': rdDistGeom.ETKDGv2(),
        'ETKDGv3': rdDistGeom.ETKDGv3(),
        'srETKDGv3': rdDistGeom.srETKDGv3(),
        'KDG': rdDistGeom.KDG()
    }
    mol = Mol(mol)
    embedding_method = available_embedding_methods.get(method)
    if embedding_method is None:
        warnings.warn(
            f'{method} is not available. Proceeding with ETKDGv3.'
        )
        embedding_method = available_embedding_methods['ETKDGv3']

    for key, value in kwargs.items():
        setattr(embedding_method, key, value)

    if not timeout:
        timeout = 0 # No timeout

    if not random_seed:
        random_seed = -1 # No random seed

    embedding_method.randomSeed = random_seed
    embedding_method.timeout = timeout

    success = rdDistGeom.EmbedMultipleConfs(
        mol, numConfs=num_conformers, params=embedding_method
    )
    num_successes = len(success)
    if num_successes < num_conformers:
        warnings.warn(
            f'Could only embed {num_successes} out of {num_conformers} conformer(s) for '
            f'{mol} using the specified method ({method}) and parameters. Attempting to '
            f'embed the remaining {num_conformers-num_successes} using fallback methods.',
        )
        max_iters = 20 * mol.num_atoms # Doubling the number of iterations
        for fallback_method in [method, 'ETDG', 'KDG']:
            fallback_embedding_method = available_embedding_methods[fallback_method]
            fallback_embedding_method.useRandomCoords = True
            fallback_embedding_method.maxIterations = int(max_iters)
            fallback_embedding_method.clearConfs = False
            fallback_embedding_method.timeout = int(timeout)
            fallback_embedding_method.randomSeed = int(random_seed)
            success = rdDistGeom.EmbedMultipleConfs(
                mol, numConfs=(num_conformers - num_successes), params=fallback_embedding_method
            )
            num_successes += len(success)
            if num_successes == num_conformers:
                break
        else:
            raise RuntimeError(
                f'Could not embed {num_conformers} conformer(s) for {mol}. '
            )
    return mol

def optimize_conformers(
    mol: Mol,
    method: str = 'UFF',
    max_iter: int = 200, 
    num_threads: bool = 1,
    ignore_interfragment_interactions: bool = True,
    vdw_threshold: float = 10.0,
) -> Mol:
    if mol.num_conformers == 0:
        warnings.warn(
            f'{mol} has no conformers to optimize. Proceeding without it.'
        )
        return Mol(mol)
    available_force_field_methods = ['MMFF', 'MMFF94', 'MMFF94s', 'UFF']
    if method not in available_force_field_methods:
        warnings.warn(
            f'{method} is not available. Proceeding with universal force field (UFF).'
        )
        method = 'UFF'
    mol_optimized = Mol(mol)
    try:
        if method.startswith('MMFF'):
            variant = method 
            if variant == 'MMFF':
                variant += '94'
            _, _ = _mmff_optimize_conformers(
                mol_optimized, 
                num_threads=num_threads, 
                max_iter=max_iter, 
                variant=variant,
                ignore_interfragment_interactions=ignore_interfragment_interactions,
            )
        else:
            _, _ = _uff_optimize_conformers(
                mol_optimized,
                num_threads=num_threads,
                max_iter=max_iter,
                vdw_threshold=vdw_threshold,
                ignore_interfragment_interactions=ignore_interfragment_interactions,
            )
    except RuntimeError as e:
        warnings.warn(
            f'Unsuccessful {method} force field minimization for {mol}. Proceeding without it.',
        )
        return Mol(mol)
    return mol_optimized

def prune_conformers(
    mol: Mol,
    keep: int = 1,
    threshold: float = 0.0,
    energy_force_field: str = 'UFF',
) -> Mol:
    if mol.num_conformers == 0:
        warnings.warn(
            f'{mol} has no conformers to prune. Proceeding without it.'
        )
        return RDKitMol(mol)
    
    threshold = threshold or 0.0
    deviations = conformer_deviations(mol)
    energies = conformer_energies(mol, method=energy_force_field)
    sorted_indices = np.argsort(energies)

    selected = [int(sorted_indices[0])]

    for target in sorted_indices[1:]:
        if len(selected) >= keep:
            break 
        if np.all(deviations[target, selected] >= threshold):
            selected.append(int(target))

    mol_copy = Mol(mol)
    mol_copy.RemoveAllConformers()
    for cid in selected:
        conformer = mol.get_conformer(cid)
        mol_copy.AddConformer(conformer, assignId=True)

    return mol_copy

def _uff_optimize_conformers(
    mol: Mol,
    num_threads: int = 1,
    max_iter: int = 200,
    vdw_threshold: float = 10.0,
    ignore_interfragment_interactions: bool = True,
    **kwargs,
) -> tuple[list[float], list[bool]]:
    """Universal Force Field Minimization.
    """
    results = rdForceFieldHelpers.UFFOptimizeMoleculeConfs(
        mol,
        numThreads=num_threads,
        maxIters=max_iter,
        vdwThresh=vdw_threshold,
        ignoreInterfragInteractions=ignore_interfragment_interactions,
    )
    energies = [r[1] for r in results]
    converged = [r[0] == 0 for r in results]
    return energies, converged 

def _mmff_optimize_conformers(
    mol: Mol,
    num_threads: int = 1,
    max_iter: int = 200,
    variant: str = 'MMFF94',
    ignore_interfragment_interactions: bool = True,
    **kwargs,
) -> tuple[list[float], list[bool]]:
    """Merck Molecular Force Field Minimization.
    """
    if not rdForceFieldHelpers.MMFFHasAllMoleculeParams(mol):
        raise ValueError("Cannot minimize molecule using MMFF.")
    rdForceFieldHelpers.MMFFSanitizeMolecule(mol)
    results = rdForceFieldHelpers.MMFFOptimizeMoleculeConfs(
        mol,
        num_threads=num_threads,
        maxIters=max_iter,
        mmffVariant=variant,
        ignoreInterfragInteractions=ignore_interfragment_interactions,
    )
    energies = [r[1] for r in results]
    converged = [r[0] == 0 for r in results]
    return energies, converged 

def _calc_uff_energies(
    mol: Mol,
) -> list[float]:
    energies = []
    for i in range(mol.num_conformers):
        try:
            force_field = rdForceFieldHelpers.UFFGetMoleculeForceField(mol, confId=i)
            energies.append(force_field.CalcEnergy())
        except Exception:
            energies.append(float('nan'))
    return energies
    
def _calc_mmff_energies(
    mol: Mol,
    variant: str = 'MMFF94',
) -> list[float]:
    energies = []
    if not rdForceFieldHelpers.MMFFHasAllMoleculeParams(mol):
        raise ValueError("Cannot compute MMFF energies for this molecule.")
    props = rdForceFieldHelpers.MMFFGetMoleculeProperties(mol, mmffVariant=variant)
    for i in range(mol.num_conformers):
        try:
            force_field = rdForceFieldHelpers.MMFFGetMoleculeForceField(mol, props, confId=i)
            energies.append(force_field.CalcEnergy())
        except Exception:
            energies.append(float('nan'))
    return energies

def unpack_conformers(mol: Mol) -> list[Mol]:
    mols = []
    for conf in mol.get_conformers():
        new_mol = RDKitMol(mol)
        new_mol.RemoveAllConformers()
        new_mol.AddConformer(conf, assignId=True)
        new_mol.__class__ = mol.__class__
        mols.append(new_mol)
    return mols

_fingerprint_types = {
    'rdkit': rdFingerprintGenerator.GetRDKitFPGenerator,
    'morgan': rdFingerprintGenerator.GetMorganGenerator,
    'topological_torsion': rdFingerprintGenerator.GetTopologicalTorsionGenerator,
    'atom_pair': rdFingerprintGenerator.GetAtomPairGenerator,
}

def _get_fingerprint(
    mol: Mol,
    fp_type: str = 'morgan',
    binary: bool = True,
    dtype: str = 'float32',
    **kwargs,
) -> np.ndarray:
    fingerprint: rdFingerprintGenerator.FingerprintGenerator64 = (
        _fingerprint_types[fp_type](**kwargs)
    )
    if not isinstance(mol, Mol):
        mol = Mol.from_encoding(mol)
    if binary:
        fp: np.ndarray = fingerprint.GetFingerprintAsNumPy(mol)
    else:
        fp: np.ndarray = fingerprint.GetCountFingerprintAsNumPy(mol)
    return fp.astype(dtype)

def _rdkit_fingerprint(
    mol: RDKitMol, 
    size: int = 2048, 
    *,
    min_path: int = 1, 
    max_path: int = 7, 
    binary: bool = True,
    dtype: str = 'float32',
) -> np.ndarray:
    fp_param = {'fpSize': size, 'minPath': min_path, 'maxPath': max_path}
    return _get_fingerprint(mol, 'rdkit', binary, dtype, **fp_param)

def _morgan_fingerprint(
    mol: RDKitMol, 
    size: int = 2048, 
    *,
    radius: int = 3, 
    binary: bool = True,
    dtype: str = 'float32',
) -> np.ndarray:
    fp_param = {'radius': radius, 'fpSize': size}
    return _get_fingerprint(mol, 'morgan', binary, dtype, **fp_param)

def _topological_torsion_fingerprint(
    mol: RDKitMol, 
    size: int = 2048, 
    *,
    binary: bool = True,
    dtype: str = 'float32',
) -> np.ndarray:
    fp_param = {'fpSize': size}
    return _get_fingerprint(mol, 'topological_torsion', binary, dtype, **fp_param)

def _atom_pair_fingerprint(
    mol: RDKitMol, 
    size: int = 2048, 
    *,
    binary: bool = True,
    dtype: str = 'float32',
) -> np.ndarray:
    fp_param = {'fpSize': size}
    return _get_fingerprint(mol, 'atom_pair', binary, dtype, **fp_param)




import keras
import warnings

from molcraft import layers
from molcraft import models
from molcraft import tensors 


@keras.saving.register_keras_serializable(package='molcraft')
class Trainer(models.GraphModel):

    '''Base trainer.

    Wraps and (pre)trains a graph neural network for a certain task.

    Args:
        model: 
            A `models.GraphModel` to be (pre)trained.
    '''
    
    def __init__(self, model: models.GraphModel, **kwargs) -> None:
        super().__init__(**kwargs)
        self.model = model 
    
    def get_config(self) -> dict:
        config = super().get_config()
        config['model'] = keras.saving.serialize_keras_object(self.model)
        return config
    
    @classmethod
    def from_config(cls, config: dict) -> 'Trainer':
        config['model'] = keras.saving.deserialize_keras_object(config['model'])
        return super().from_config(config)
    

@keras.saving.register_keras_serializable(package='molcraft')
class NodePredictionTrainer(Trainer):

    '''Node prediction trainer.

    Wraps and (pre)trains a graph neural network to perform node predictions.

    Ignores super nodes and edges, if they exist.

    Args:
        model: 
            A `models.GraphModel` to be (pre)trained.
        decoder:
            An optional decoder for converting updated node features to node predictions.
            If None, a `keras.layers.Dense` layer is used with `units` set to `label` dim.
        select_rate:
            The rate of which nodes will be selected for prediction. If None, all nodes are predicted.
        mask_selected:
            Whether to mask the selected nodes. Only relevant if `select_rate` is specified.
        edge_masking_rate:
            The rate of which edges will be masked. If None, edges will not be masked.
            Only relevant if `select_rate` is specified.

    Example:

    >>> import molcraft
    >>> import keras
    >>> 
    >>> featurizer = molcraft.featurizers.MolGraphFeaturizer(
    ...     atom_features=[
    ...         molcraft.features.AtomType(['C', 'N', 'O', 'P', 'S']),
    ...     ]
    ... )
    >>> graph = featurizer(['N[C@@H](C)C(=O)O', 'N[C@@H](CS)C(=O)O'])
    >>> # Label nodes with the one-hot encoded atom types for illustration
    >>> graph = graph.update({'node': {'label': graph.node['feature']}})
    >>> 
    >>> inputs = molcraft.layers.Input(graph.spec)
    >>> x = molcraft.layers.NodeEmbedding(128)(inputs)
    >>> x = molcraft.layers.EdgeEmbedding(128)(x)
    >>> x = molcraft.layers.GraphConv(128)(x)
    >>> outputs = molcraft.layers.GraphConv(128)(x)
    >>> model = molcraft.models.GraphModel(inputs, outputs)
    >>> 
    >>> pretrainer = molcraft.trainers.NodePredictionTrainer(
    ...     model,
    ...     decoder=None,       # Dense(units=node_label_dim)
    ...     select_rate=0.5,
    ...     mask_selected=True,
    ... )
    >>> pretrainer.compile(
    ...     optimizer=keras.optimizers.Adam(1e-4),
    ...     loss=keras.losses.CategoricalCrossentropy(from_logits=True),
    ... )
    >>> pretrainer.fit(graph, epochs=10)
    >>> # pretrainer.model.save('/tmp/model.keras')
    '''

    def __init__(
        self,
        model: models.GraphModel,
        decoder: keras.layers.Layer | None = None,
        select_rate: float | None = None,
        mask_selected: bool = False,
        edge_masking_rate: float | None = None,
        **kwargs
    ) -> None:
        super().__init__(model=model, **kwargs)
        
        for layer in self.model.layers:
            if isinstance(layer, layers.NodeEmbedding):
                break
        else:
            raise ValueError('Could not find `NodeEmbedding` layer.')
        
        self._embedder = models.GraphModel(
            self.model.input, layer._symbolic_output
        )
        self._model = models.GraphModel(
            layer._symbolic_output, self.model.output
        )
        self._decoder = decoder
        self._select_rate = select_rate
        self._mask_selected = mask_selected
        if edge_masking_rate and not mask_selected:
            warnings.warn(
                'Setting `edge_masking_rate` to `None`, '
                'as `mask_selected` is set to `False`.'
            )
            edge_masking_rate = None
        self._edge_masking_rate = edge_masking_rate
    
    def build(self, spec: tensors.GraphTensor.Spec) -> None:

        self._has_super = ('super' in spec.node)
        self._has_edge_feature = ('feature' in spec.edge)

        if self._mask_selected:
            node_feature_dim = self._embedder._symbolic_output['node']['feature'].shape[-1]
            self._node_mask_feature = self.get_weight(shape=[node_feature_dim])

        if self._mask_selected and self._has_edge_feature and self._edge_masking_rate:
            edge_feature_dim = self._embedder._symbolic_output['edge']['feature'].shape[-1]
            self._edge_mask_feature = self.get_weight(shape=[edge_feature_dim])
        elif self._edge_masking_rate and not self._has_edge_feature:
            warnings.warn(
                'Setting `edge_masking_rate` to `None`, '
                'as no edge features exist.'
            )
            self._edge_masking_rate = None

        if self._decoder is None:
            label_dim = spec.node['label'].shape[-1]
            self._decoder = keras.layers.Dense(units=label_dim) 

    def propagate(
        self, 
        tensor: tensors.GraphTensor,
        training: bool | None = None,
    ) -> tensors.GraphTensor:
        sample_weight = tensor.node.get('sample_weight')
        if sample_weight is None:
            sample_weight = keras.ops.ones([tensor.num_nodes])

        tensor = self._embedder(tensor)

        if self._select_rate is not None and training:
            # Select nodes to be predicted
            is_not_super = (
                True if not self._has_super else keras.ops.logical_not(tensor.node['super'])
            )
            r = keras.random.uniform(shape=[tensor.num_nodes])
            node_mask = keras.ops.logical_and(is_not_super, self._select_rate > r)
            sample_weight = keras.ops.where(node_mask, sample_weight, 0.0)

            if self._mask_selected:
                # Mask selected node features
                node_feature_masked = keras.ops.where(
                    node_mask[:, None], self._node_mask_feature, tensor.node['feature']
                )
                tensor = tensor.update({'node': {'feature': node_feature_masked}})

                if self._edge_masking_rate:
                    # Mask edge features
                    is_not_super = (
                        True if not self._has_super else keras.ops.logical_not(tensor.edge['super'])
                    )
                    r = keras.random.uniform(shape=[tensor.num_edges])
                    edge_mask = keras.ops.logical_and(is_not_super, self._edge_masking_rate > r)
                    edge_feature_masked = keras.ops.where(
                        edge_mask[:, None], self._edge_mask_feature, tensor.edge['feature']
                    )
                    tensor = tensor.update({'edge': {'feature': edge_feature_masked}})

        node_feature = self._model(tensor).node['feature']
        node_prediction = self._decoder(node_feature)
        return tensor.update({
            'node': {
                'prediction': node_prediction, 
                'sample_weight': sample_weight
            }
        })
    
    def get_config(self) -> dict:
        config = super().get_config()
        config.update({
            'decoder': keras.saving.serialize_keras_object(self._decoder),
            'select_rate': self._select_rate,
            'mask_selected': self._mask_selected,
            'edge_masking_rate': self._edge_masking_rate,
        })
        return config
    
    @classmethod
    def from_config(cls, config: dict) -> 'NodePredictionTrainer':
        config['decoder'] = keras.saving.deserialize_keras_object(config['decoder'])
        return super().from_config(config)


import unittest 
import numpy as np

from molcraft import chem


class TestChem(unittest.TestCase):

    def setUp(self):
        self.smiles = [
            "N[C@@H](CC(=O)N)C(=O)O",
            "N1[C@@H](CCC1)C(=O)O",
        ] 


if __name__ == '__main__':
    unittest.main()


import warnings
import keras
import numpy as np


class TensorBoard(keras.callbacks.TensorBoard):

    def _log_weights(self, epoch):
        with self._train_writer.as_default():
            for layer in self.model.layers:
                for weight in layer.weights:
                    # Use weight.path istead of weight.name to distinguish
                    # weights of different layers.
                    histogram_weight_name = weight.path + "/histogram"
                    self.summary.histogram(
                        histogram_weight_name, weight, step=epoch
                    )
                    if self.write_images:
                        image_weight_name = weight.path + "/image"
                        self._log_weight_as_image(
                            weight, image_weight_name, epoch
                        )
            self._train_writer.flush()


class LearningRateDecay(keras.callbacks.LearningRateScheduler):

    def __init__(self, rate: float, delay: int = 0, **kwargs):

        def lr_schedule(epoch: int, lr: float):
            if epoch < delay:
                return float(lr)
            return float(lr * keras.ops.exp(-rate))
        
        super().__init__(schedule=lr_schedule, **kwargs)


class Rollback(keras.callbacks.Callback):
    """Rollback callback.

    Currently, this callback simply restores the model and (optionally) the optimizer 
    variables if current loss deviates too much from the best observed loss.

    This callback might be useful in situations where the loss tend to spike and put 
    the model in an undesired/problematic high-loss parameter space.

    Args:
        tolerance (float):
            The threshold for when the restoration is triggered. The devaiation is 
            calculated as follows: (current_loss - best_loss) / best_loss.
    """

    def __init__(
        self,
        tolerance: float = 0.5,
        rollback_optimizer: bool = True,
    ):
        super().__init__()
        self.tolerance = tolerance
        self.rollback_optimizer = rollback_optimizer

    def on_train_begin(self, logs=None):
        self._rollback_weights = self._get_model_vars()
        if self.rollback_optimizer:
            self._rollback_optimizer_vars = self._get_optimizer_vars()
        self._rollback_loss = float('inf')

    def on_epoch_end(self, epoch: int, logs: dict = None):
        current_loss = logs.get('val_loss', logs.get('loss'))
        deviation = (current_loss - self._rollback_loss) / self._rollback_loss

        if np.isnan(current_loss) or np.isinf(current_loss):
            self._rollback()
            # Rolling back model because of nan or inf loss
            return

        if deviation > self.tolerance:
            self._rollback()
            # Rolling back model because of large loss deviation.
            return

        if current_loss < self._rollback_loss:
            self._save_state(current_loss)

    def _save_state(self, current_loss: float) -> None:
        self._rollback_loss = current_loss
        self._rollback_weights = self._get_model_vars()
        if self.rollback_optimizer:
            self._rollback_optimizer_vars = self._get_optimizer_vars()

    def _rollback(self) -> None:
        self.model.set_weights(self._rollback_weights)
        if self.rollback_optimizer:
            self.model.optimizer.set_weights(self._rollback_optimizer_vars)

    def _get_optimizer_vars(self):
        return [v.numpy() for v in self.model.optimizer.variables]

    def _get_model_vars(self):
        return self.model.get_weights()



import unittest 

import keras
import tensorflow as tf

from molcraft import tensors 


class TestGraphTensor(unittest.TestCase):

    def setUp(self):
        self.flat_tensors = [
            # Graph with two subgraphs
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([2, 3], dtype='int32')
                },
                node={
                    'feature': keras.ops.array([[1., 2.], [3., 4.], [5., 6.], [6., 7.], [8., 9.]], dtype='float32'),
                    'weight': keras.ops.array([0.50, 1.00, 2.00, 0.25, 0.75], dtype='float32')
                },
                edge={
                    'source': keras.ops.array([0, 1, 3, 4, 4, 3, 2], dtype='int32'),
                    'target': keras.ops.array([1, 0, 2, 3, 4, 4, 3], dtype='int32'),
                    'feature': keras.ops.array([[1.], [2.], [3.], [4.], [5.], [6.], [7.]], dtype='float32')
                }
            ),
            # Graph with two subgraphs, none which has edges
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([2, 3], dtype='int32')
                },
                node={
                    'feature': keras.ops.array([[1., 2.], [3., 4.], [5., 6.], [6., 7.], [8., 9.]], dtype='float32'),
                    'weight': keras.ops.array([0.50, 1.00, 2.00, 0.25, 0.75], dtype='float32')
                },
                edge={
                    'source': keras.ops.zeros([0], dtype='int32'),
                    'target': keras.ops.zeros([0], dtype='int32'),
                    'feature': keras.ops.zeros([0, 1], dtype='float32')
                }
            ),
            # Graph with two subgraphs, none of which has nodes or edges
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([0, 0], dtype='int32')
                },
                node={
                    'feature': keras.ops.zeros([0, 2], dtype='float32'),
                    'weight': keras.ops.zeros([0], dtype='float32')
                },
                edge={
                    'source': keras.ops.zeros([0], dtype='int32'),
                    'target': keras.ops.zeros([0], dtype='int32'),
                    'feature': keras.ops.zeros([0, 1], dtype='float32')
                }
            ),
            # Graph with three subgraphs where the second subgraph's first node has no edges
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([2, 3, 2], dtype='int32')
                },
                node={
                    'feature': keras.ops.array([[1., 2.], [3., 4.], [5., 6.], [6., 7.], [8., 9.], [10., 11.], [12., 13.]], dtype='float32'),
                    'weight': keras.ops.array([0.50, 1.00, 2.00, 0.25, 0.75, 0.25, 0.25], dtype='float32')
                },
                edge={
                    'source': keras.ops.array([0, 1, 4, 3, 4, 5, 6, 6], dtype='int32'),
                    'target': keras.ops.array([1, 0, 3, 4, 4, 6, 5, 6], dtype='int32'),
                    'feature': keras.ops.array([[1.], [2.], [3.], [4.], [5.], [6.], [7.], [8.]], dtype='float32')
                }
            ),
            # Graph with three subgraphs where the first subgraph's nodes have no edges
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([2, 3, 2], dtype='int32')
                },
                node={
                    'feature': keras.ops.array([[1., 2.], [3., 4.], [5., 6.], [6., 7.], [8., 9.], [10., 11.], [12., 13.]], dtype='float32'),
                    'weight': keras.ops.array([0.50, 1.00, 2.00, 0.25, 0.75, 0.25, 0.25], dtype='float32')
                },
                edge={
                    'source': keras.ops.array([4, 3, 4, 5, 6, 6], dtype='int32'),
                    'target': keras.ops.array([3, 4, 4, 6, 5, 6], dtype='int32'),
                    'feature': keras.ops.array([[3.], [4.], [5.], [6.], [7.], [8.]], dtype='float32')
                }
            ),
            # Graph with three subgraphs where the last subgraph's nodes have no edges
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([2, 3, 3], dtype='int32')
                },
                node={
                    'feature': keras.ops.array([[1., 2.], [3., 4.], [5., 6.], [6., 7.], [8., 9.], [10., 11.], [12., 13.], [14., 15.]], dtype='float32'),
                    'weight': keras.ops.array([0.50, 1.00, 2.00, 0.25, 0.75, 0.25, 0.25, 0.5], dtype='float32')
                },
                edge={
                    'source': keras.ops.array([0, 1, 4, 3, 4, 5, 6], dtype='int32'),
                    'target': keras.ops.array([1, 0, 3, 4, 4, 6, 5], dtype='int32'),
                    'feature': keras.ops.array([[1.], [2.], [3.], [4.], [5.], [6.], [7.]], dtype='float32')
                }
            ),
        ]

        self.unflat_tensors = [
            # Graph with two subgraphs
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([2, 3], dtype='int32')
                },
                node={
                    'feature': tf.ragged.constant([[[1., 2.], [3., 4.]], [[5., 6.], [6., 7.], [8., 9.]]], dtype='float32', ragged_rank=1, row_splits_dtype='int32'),
                    'weight': tf.ragged.constant([[0.50, 1.00], [2.00, 0.25, 0.75]], dtype='float32', ragged_rank=1, row_splits_dtype='int32')
                },
                edge={
                    'source': tf.ragged.constant([[0, 1], [1, 2, 2, 1, 0]], dtype='int32', ragged_rank=1, row_splits_dtype='int32'),
                    'target': tf.ragged.constant([[1, 0], [0, 1, 2, 2, 1]], dtype='int32', ragged_rank=1, row_splits_dtype='int32'),
                    'feature': tf.ragged.constant([[[1.], [2.]], [[3.], [4.], [5.], [6.], [7.]]], dtype='float32', ragged_rank=1, row_splits_dtype='int32')
                }
            ),
            # Graph with two subgraphs, none which has edges
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([2, 3], dtype='int32')
                },
                node={
                    'feature': tf.ragged.constant([[[1., 2.], [3., 4.]], [[5., 6.], [6., 7.], [8., 9.]]], dtype='float32', ragged_rank=1, row_splits_dtype='int32'),
                    'weight': tf.ragged.constant([[0.50, 1.00], [2.00, 0.25, 0.75]], dtype='float32', ragged_rank=1, row_splits_dtype='int32')
                },
                edge={
                    'source': tf.ragged.constant([[], []], dtype='int32', ragged_rank=1, row_splits_dtype='int32'),
                    'target': tf.ragged.constant([[], []], dtype='int32', ragged_rank=1, row_splits_dtype='int32'),
                    'feature': tf.RaggedTensor.from_value_rowids(keras.ops.zeros([0, 1], dtype='float32'), keras.ops.zeros([0], dtype='int32'), 2),
                }
            ),
            # Graph with two subgraphs, none of which has nodes or edges
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([0, 0], dtype='int32')
                },
                node={
                    'feature': tf.RaggedTensor.from_value_rowids(keras.ops.zeros([0, 2], dtype='float32'), keras.ops.zeros([0], dtype='int32'), 2),
                    'weight': tf.RaggedTensor.from_value_rowids(keras.ops.zeros([0], dtype='float32'), keras.ops.zeros([0], dtype='int32'), 2)
                },
                edge={
                    'source': tf.ragged.constant([[], []], dtype='int32', ragged_rank=1, row_splits_dtype='int32'),
                    'target': tf.ragged.constant([[], []], dtype='int32', ragged_rank=1, row_splits_dtype='int32'),
                    'feature': tf.RaggedTensor.from_value_rowids(keras.ops.zeros([0, 1], dtype='float32'), keras.ops.zeros([0], dtype='int32'), 2)
                }
            ),
            # Graph with three subgraphs where the second subgraph's first node has no edges
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([2, 3, 2], dtype='int32')
                },
                node={
                    'feature': tf.ragged.constant([[[1., 2.], [3., 4.]], [[5., 6.], [6., 7.], [8., 9.]], [[10., 11.], [12., 13.]]], dtype='float32', ragged_rank=1, row_splits_dtype='int32'),
                    'weight': tf.ragged.constant([[0.50, 1.00], [2.00, 0.25, 0.75], [0.25, 0.25]], dtype='float32', ragged_rank=1, row_splits_dtype='int32')
                },
                edge={
                    'source': tf.ragged.constant([[0, 1], [2, 1, 2], [0, 1, 1]], dtype='int32', ragged_rank=1, row_splits_dtype='int32'),
                    'target': tf.ragged.constant([[1, 0], [1, 2, 2], [1, 0, 1]], dtype='int32', ragged_rank=1, row_splits_dtype='int32'),
                    'feature': tf.ragged.constant([[[1.], [2.]], [[3.], [4.], [5.]], [[6.], [7.], [8.]]], dtype='float32', ragged_rank=1, row_splits_dtype='int32')
                }
            ),
            # Graph with three subgraphs where the first subgraph's nodes have no edges
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([2, 3, 2], dtype='int32')
                },
                node={
                    'feature': tf.ragged.constant([[[1., 2.], [3., 4.]], [[5., 6.], [6., 7.], [8., 9.]], [[10., 11.], [12., 13.]]], dtype='float32', ragged_rank=1, row_splits_dtype='int32'),
                    'weight': tf.ragged.constant([[0.50, 1.00], [2.00, 0.25, 0.75], [0.25, 0.25]], dtype='float32', ragged_rank=1, row_splits_dtype='int32')
                },
                edge={
                    'source': tf.ragged.constant([[], [2, 1, 2], [0, 1, 1]], dtype='int32', ragged_rank=1, row_splits_dtype='int32'),
                    'target': tf.ragged.constant([[], [1, 2, 2], [1, 0, 1]], dtype='int32', ragged_rank=1, row_splits_dtype='int32'),
                    'feature': tf.ragged.constant([[], [[3.], [4.], [5.]], [[6.], [7.], [8.]]], dtype='float32', ragged_rank=1, row_splits_dtype='int32')
                }
            ),
            # Graph with three subgraphs where the last subgraph's nodes have no edges
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([2, 3, 3], dtype='int32')
                },
                node={
                    'feature': tf.ragged.constant([[[1., 2.], [3., 4.]], [[5., 6.], [6., 7.], [8., 9.]], [[10., 11.], [12., 13.], [14., 15.]]], dtype='float32', ragged_rank=1, row_splits_dtype='int32'),
                    'weight': tf.ragged.constant([[0.50, 1.00], [2.00, 0.25, 0.75], [0.25, 0.25, 0.5]], dtype='float32', ragged_rank=1, row_splits_dtype='int32')
                },
                edge={
                    'source': tf.ragged.constant([[0, 1], [2, 1, 2], [0, 1]], dtype='int32', ragged_rank=1, row_splits_dtype='int32'),
                    'target': tf.ragged.constant([[1, 0], [1, 2, 2], [1, 0]], dtype='int32', ragged_rank=1, row_splits_dtype='int32'),
                    'feature': tf.ragged.constant([[[1.], [2.]], [[3.], [4.], [5.]], [[6.], [7.]]], dtype='float32', ragged_rank=1, row_splits_dtype='int32')
                }
            ),
        ]

    def test_unflatten(self):
        for i, (unflat_tensor, flat_tensor) in enumerate(zip(self.unflat_tensors, self.flat_tensors)):
            with self.subTest(id=i):
                self.assertTrue(bool(unflat_tensor == flat_tensor.unflatten()))
            

    def test_flatten(self):
        for i, (unflat_tensor, flat_tensor) in enumerate(zip(self.unflat_tensors, self.flat_tensors)):
            with self.subTest(i=i):
                self.assertTrue(bool(unflat_tensor.flatten() == flat_tensor))

    def test_tf_dataset(self):
        for i, tensor in enumerate(self.flat_tensors):
            with self.subTest(i=i, flat=True):
                ds = tf.data.Dataset.from_tensor_slices(tensor)
                for x in ds.batch(2).unbatch().batch(2).take(1):
                    pass
                self.assertTrue(x == tensor[:2])

        for i, tensor in enumerate(self.unflat_tensors):
            with self.subTest(i=i, flat=False):
                ds = tf.data.Dataset.from_tensor_slices(tensor)
                for x in ds.batch(2).unbatch().batch(2).take(1):
                    pass
                self.assertTrue(x == tensor[:2])


if __name__ == '__main__':
    unittest.main()


import warnings
import collections
import functools
import inspect
import keras 
import json
import abc
import typing 
import numpy as np
import pandas as pd
import tensorflow as tf
import multiprocessing as mp

from pathlib import Path

from molcraft import tensors 
from molcraft import features
from molcraft import records
from molcraft import chem
from molcraft import descriptors


@keras.saving.register_keras_serializable(package='molcraft')
class GraphFeaturizer(abc.ABC):

    """Base graph featurizer.
    """

    @abc.abstractmethod
    def call(self, x: typing.Any, context: dict) -> tensors.GraphTensor:
        pass

    def get_config(self) -> dict:
        return {}
    
    @classmethod
    def from_config(cls, config: dict) -> 'GraphFeaturizer':
        return cls(**config)
    
    def save(self, filepath: str | Path, *args, **kwargs) -> None:
        save_featurizer(self, filepath, *args, **kwargs)

    @staticmethod
    def load(filepath: str | Path, *args, **kwargs) -> 'GraphFeaturizer':
        return load_featurizer(filepath, *args, **kwargs)
    
    def write_records(self, inputs: typing.Iterable, path: str | Path, **kwargs) -> None:
         records.write(
            inputs, featurizer=self, path=path, **kwargs
         )

    @staticmethod
    def read_records(path: str | Path, **kwargs) -> tf.data.Dataset:
        return records.read(
            path=path, **kwargs
        )

    def _call(self, inputs: typing.Any) -> tensors.GraphTensor:
        inputs, context = _unpack_inputs(inputs)
        if _call_kwargs(self.call):
            graph = self.call(inputs, context=context)
        else:
            graph = self.call(inputs)
        if not isinstance(graph, tensors.GraphTensor):
            graph = tensors.from_dict(graph)
        return graph

    def __call__(
        self,
        inputs: typing.Iterable,
        *,
        multiprocessing: bool = False,
        processes: int | None = None,
        device: str = '/cpu:0',
    ) -> tensors.GraphTensor:
        if not isinstance(
            inputs, (list, np.ndarray, pd.Series, pd.DataFrame, typing.Generator)
        ):
            return self._call(inputs)
        
        if isinstance(inputs, (np.ndarray, pd.Series)):
            inputs = inputs.tolist()
        elif isinstance(inputs, pd.DataFrame):
            inputs = inputs.iterrows()

        if not multiprocessing:
            outputs = [self._call(x) for x in inputs]
        else:
            with tf.device(device):
                with mp.Pool(processes) as pool:
                    outputs = pool.map(func=self._call, iterable=inputs)
        outputs = [x for x in outputs if x is not None]
        if tensors.is_scalar(outputs[0]):
            return tf.stack(outputs, axis=0)
        return tf.concat(outputs, axis=0)


@keras.saving.register_keras_serializable(package='molcraft')
class MolGraphFeaturizer(GraphFeaturizer):

    """Molecular graph featurizer.

    Converts SMILES or InChI strings to a molecular graph.

    The molecular graph may encode a single molecule or a batch of molecules.

    Example:

    >>> import molcraft 
    >>> 
    >>> featurizer = molcraft.featurizers.MolGraphFeaturizer(
    ...     atom_features=[
    ...         molcraft.features.AtomType(),
    ...         molcraft.features.NumHydrogens(),
    ...         molcraft.features.Degree(),
    ...     ],
    ...     bond_features=[
    ...         molcraft.features.BondType(),
    ...     ],
    ...     super_node=False,
    ...     self_loops=False,
    ... )
    >>> 
    >>> graph = featurizer(["N[C@@H](C)C(=O)O", "N[C@@H](CS)C(=O)O"])
    >>> graph
    GraphTensor(
        context={
            'size': <tf.Tensor: shape=[2], dtype=int32>
        },
        node={
            'feature': <tf.Tensor: shape=[13, 129], dtype=float32>
        },
        edge={
            'source': <tf.Tensor: shape=[22], dtype=int32>,
            'target': <tf.Tensor: shape=[22], dtype=int32>,
            'feature': <tf.Tensor: shape=[22, 5], dtype=float32>
        }
    )

    Args:
        atom_features:
            A list of `features.Feature` encoded as the node features.
        bond_features:
            A list of `features.Feature` encoded as the edge features.
        molecule_features:
            A list of `descriptors.Descriptor` encoded as the context feature.
        super_node:
            A boolean specifying whether to include a super node.
        self_loops:
            A boolean specifying whether self loops exist.
        include_hydrogens:
            A boolean specifying whether hydrogens should be encoded as nodes.
        wildcards:
            A boolean specifying whether wildcards exist. If True, wildcard labels will
            be encoded in the graph and separately embedded in `layers.NodeEmbedding`.
    """

    def __init__(
        self,
        atom_features: list[features.Feature] | str = 'auto',
        bond_features: list[features.Feature] | str | None = 'auto',
        molecule_features: list[descriptors.Descriptor] | str | None = None,
        super_node: bool = False,
        self_loops: bool = False,
        include_hydrogens: bool = False,
        wildcards: bool = False,
    ) -> None:
        use_default_atom_features = (
            atom_features == 'auto' or atom_features == 'default'
        )
        if use_default_atom_features:
            atom_features = [features.AtomType(), features.Degree()]
            if not include_hydrogens:
                atom_features += [features.NumHydrogens()]

        use_default_bond_features = (
            bond_features == 'auto' or bond_features == 'default'
        )
        if use_default_bond_features:
            bond_features = [features.BondType()]

        use_default_molecule_features = (
            molecule_features == 'auto' or molecule_features == 'default'
        )
        if use_default_molecule_features:
            molecule_features = [
                descriptors.MolWeight(),
                descriptors.TotalPolarSurfaceArea(),
                descriptors.LogP(),
                descriptors.MolarRefractivity(),
                descriptors.NumHeavyAtoms(),
                descriptors.NumHeteroatoms(),
                descriptors.NumHydrogenDonors(),
                descriptors.NumHydrogenAcceptors(),
                descriptors.NumRotatableBonds(),
                descriptors.NumRings(),
            ]

        self._atom_features = atom_features
        self._bond_features = bond_features
        self._molecule_features = molecule_features
        self._include_hydrogens = include_hydrogens
        self._wildcards = wildcards
        self._self_loops = self_loops
        self._super_node = super_node

    def call(
        self, 
        mol: str | chem.Mol | tuple, 
        context: dict | None = None
    ) -> tensors.GraphTensor:

        if isinstance(mol, str):
            mol = chem.Mol.from_encoding(
                mol, explicit_hs=self._include_hydrogens
            )
        elif isinstance(mol, chem.RDKitMol):
            mol = chem.Mol.cast(mol)

        data = {'context': {}, 'node': {}, 'edge': {}}
        
        data['context']['size'] = np.asarray(mol.num_atoms)

        if self._molecule_features is not None:
            data['context']['feature'] = np.concatenate(
                [f(mol) for f in self._molecule_features], axis=-1
            )

        if context:
            data['context'].update(context)

        data['node']['feature'] = np.concatenate(
            [f(mol) for f in self._atom_features], axis=-1
        )

        if self._wildcards:
            wildcard_labels = np.asarray([
                (atom.label or 0) + 1 if atom.symbol == "*" else 0
                for atom in mol.atoms
            ])
            data['node']['wildcard'] = wildcard_labels
            data['node']['feature'] = np.where(
                wildcard_labels[:, None],
                np.zeros_like(data['node']['feature']),
                data['node']['feature']
            )

        data['edge']['source'], data['edge']['target'] = mol.adjacency(
            fill='full', sparse=True, self_loops=self._self_loops
        )
        
        if self._bond_features is not None:
            bond_features = np.concatenate(
                [f(mol) for f in self._bond_features], axis=-1
            )
            if self._self_loops:
                bond_features = np.pad(bond_features, [(0, 1), (0, 0)])

            bond_indices = [
                mol.get_bond_between_atoms(i, j).index if (i != j) else -1
                for (i, j) in zip(data['edge']['source'], data['edge']['target'])
            ]

            data['edge']['feature'] = bond_features[bond_indices]

        if self._super_node:
            data = _add_super_node(data)

        return tensors.GraphTensor(**_convert_dtypes(data))
    
    def get_config(self):
        config = super().get_config()
        config.update({
            'atom_features': keras.saving.serialize_keras_object(
                self._atom_features
            ),
            'bond_features': keras.saving.serialize_keras_object(
                self._bond_features
            ),
            'molecule_features': keras.saving.serialize_keras_object(
                self._molecule_features
            ),
            'super_node': self._super_node,
            'self_loops': self._self_loops,
            'include_hydrogens': self._include_hydrogens,
            'wildcards': self._wildcards,
        })
        return config

    @classmethod
    def from_config(cls, config: dict):
        config['atom_features'] = keras.saving.deserialize_keras_object(
            config['atom_features']
        )
        config['bond_features'] = keras.saving.deserialize_keras_object(
            config['bond_features']
        )
        config['molecule_features'] = keras.saving.deserialize_keras_object(
            config['molecule_features']
        )
        return cls(**config)
    

@keras.saving.register_keras_serializable(package='molcraft')
class MolGraphFeaturizer3D(MolGraphFeaturizer):

    """3D Molecular graph featurizer.

    Converts SMILES or InChI strings to a 3d molecular graph.

    The molecular graph may encode a single molecule or a batch of molecules.

    Example:

    >>> import molcraft 
    >>> 
    >>> featurizer = molcraft.featurizers.MolGraphFeaturizer3D(
    ...     atom_features=[
    ...         molcraft.features.AtomType(),
    ...         molcraft.features.NumHydrogens(),
    ...         molcraft.features.Degree(),
    ...     ],
    ...     radius=5.0,
    ...     random_seed=42,
    ... )
    >>> 
    >>> graph = featurizer(["N[C@@H](C)C(=O)O", "N[C@@H](CS)C(=O)O"])
    >>> graph
    GraphTensor(
        context={
            'size': <tf.Tensor: shape=[2], dtype=int32>
        },
        node={
            'feature': <tf.Tensor: shape=[13, 129], dtype=float32>,
            'coordinate': <tf.Tensor: shape=[13, 3], dtype=float32>
        },
        edge={
            'source': <tf.Tensor: shape=[72], dtype=int32>,
            'target': <tf.Tensor: shape=[72], dtype=int32>,
            'feature': <tf.Tensor: shape=[72, 12], dtype=float32>
        }
    )
        
    Args:
        atom_features:
            A list of `features.Feature` encoded as the node features.
        pair_features:
            A list of `features.PairFeature` encoded as the edge features.
        molecule_features:
            A list of `descriptors.Descriptor` encoded as the context feature.
        super_node:
            A boolean specifying whether to include a super node.
        self_loops:
            A boolean specifying whether self loops exist.
        include_hydrogens:
            A boolean specifying whether hydrogens should be encoded as nodes.
        wildcards:
            A boolean specifying whether wildcards exist. If True, wildcard labels will
            be encoded in the graph and separately embedded in `layers.NodeEmbedding`.
        radius:
            A floating point value specifying maximum edge length. 
        random_seed:
            An integer specifying the random seed for the conformer generation.
    """

    def __init__(
        self,
        atom_features: list[features.Feature] | str = 'auto',
        pair_features: list[features.PairFeature] | str = 'auto',
        molecule_features: features.Feature | str | None = None,
        super_node: bool = False,
        self_loops: bool = False,
        include_hydrogens: bool = False,
        wildcards: bool = False,
        radius: int | float | None = 6.0,
        random_seed: int | None = None,
        **kwargs,
    ) -> None:
        kwargs.pop('bond_features', None)
        super().__init__(
            atom_features=atom_features,
            bond_features=None,
            molecule_features=molecule_features,
            super_node=super_node,
            self_loops=self_loops,
            include_hydrogens=include_hydrogens,
            wildcards=wildcards,
        )

        use_default_pair_features = (
            pair_features == 'auto' or pair_features == 'default'
        )
        if use_default_pair_features:
            pair_features = [features.PairDistance()]

        self._pair_features = pair_features 
        self._radius = float(radius) if radius else None
        self._random_seed = random_seed

    def call(
        self, 
        mol: str | chem.Mol | tuple, 
        context: dict | None = None
    ) -> tensors.GraphTensor:

        if isinstance(mol, str):
            mol = chem.Mol.from_encoding(
                mol, explicit_hs=True
            )
        elif isinstance(mol, chem.RDKitMol):
            mol = chem.Mol.cast(mol)

        if mol.num_conformers == 0:
            mol = chem.embed_conformers(
                mol, num_conformers=1, random_seed=self._random_seed
            )

        if not self._include_hydrogens:
            mol = chem.remove_hs(mol)

        data = {'context': {}, 'node': {}, 'edge': {}}

        data['context']['size'] = np.asarray(mol.num_atoms)

        if self._molecule_features is not None:
            data['context']['feature'] = np.concatenate(
                [f(mol) for f in self._molecule_features], axis=-1
            )

        if context:
            data['context'].update(context)

        conformer = mol.get_conformer()

        data['node']['feature'] = np.concatenate(
            [f(mol) for f in self._atom_features], axis=-1
        )
        data['node']['coordinate'] = conformer.coordinates

        if self._wildcards:
            wildcard_labels = np.asarray([
                (atom.label or 0) + 1 if atom.symbol == "*" else 0
                for atom in mol.atoms
            ])
            data['node']['wildcard'] = wildcard_labels
            data['node']['feature'] = np.where(
                wildcard_labels[:, None],
                np.zeros_like(data['node']['feature']),
                data['node']['feature']
            )

        adjacency_matrix = conformer.adjacency(
            fill='full', radius=self._radius, sparse=False, self_loops=self._self_loops,
        )

        data['edge']['source'], data['edge']['target'] = np.where(adjacency_matrix)
        
        if self._pair_features is not None:
            pair_features = np.concatenate(
                [f(mol) for f in self._pair_features], axis=-1
            )
            pair_keep = adjacency_matrix.reshape(-1).astype(bool)
            data['edge']['feature'] = pair_features[pair_keep]

        if self._super_node:
            data = _add_super_node(data)
            data['node']['coordinate'] = np.concatenate(
                [data['node']['coordinate'], conformer.centroid[None]], axis=0
            )

        return tensors.GraphTensor(**_convert_dtypes(data))
    
    @property 
    def random_seed(self) -> int | None:
        return self._random_seed
    
    @random_seed.setter
    def random_seed(self, value: int) -> None:
        self._random_seed = value 

    def get_config(self):
        config = super().get_config()
        config['radius'] = self._radius
        config['pair_features'] = keras.saving.serialize_keras_object(
            self._pair_features
        )
        config['random_seed'] = self._random_seed
        return config

    @classmethod
    def from_config(cls, config: dict):
        config['pair_features'] = keras.saving.deserialize_keras_object(
            config['pair_features']
        )
        return super().from_config(config)


def save_featurizer(
    featurizer: GraphFeaturizer, 
    filepath: str | Path, 
    overwrite: bool = True, 
    **kwargs
) -> None:
    filepath = Path(filepath)
    if filepath.suffix != '.json':
        raise ValueError(
            'Invalid `filepath` extension for saving a `GraphFeaturizer`. '
            'A `GraphFeaturizer` should be saved as a JSON file.'
        )
    if not filepath.parent.exists():
        filepath.parent.mkdir(parents=True, exist_ok=True)
    if filepath.exists() and not overwrite:
        return 
    serialized_featurizer = keras.saving.serialize_keras_object(featurizer)
    with open(filepath, 'w') as f:
        json.dump(serialized_featurizer, f, indent=4)

def load_featurizer(
    filepath: str | Path,
    **kwargs
) -> GraphFeaturizer:
    filepath = Path(filepath)
    if filepath.suffix != '.json':
        raise ValueError(
            'Invalid `filepath` extension for loading a `GraphFeaturizer`. '
            'A `GraphFeaturizer` should be saved as a JSON file.'
        )
    if not filepath.exists():
        return 
    with open(filepath, 'r') as f:
        config = json.load(f)
    return keras.saving.deserialize_keras_object(config)

def _add_super_node(
    data: dict[str, dict[str, np.ndarray]]
) -> dict[str, dict[str, np.ndarray]]:

    data['context']['size'] += 1

    num_nodes = data['node']['feature'].shape[0]
    num_edges = data['edge']['source'].shape[0]
    super_node_index = num_nodes

    add_self_loops = np.any(
        data['edge']['source'] == data['edge']['target']
    )
    if add_self_loops:
        data['edge']['source'] = np.append(
            data['edge']['source'], super_node_index
        )
        data['edge']['target'] = np.append(
            data['edge']['target'], super_node_index
        )

    data['node']['feature'] = np.pad(data['node']['feature'], [(0, 1), (0, 0)])
    data['node']['super'] = np.asarray([False] * num_nodes + [True])
    if 'wildcard' in data['node']:
        data['node']['wildcard'] = np.pad(data['node']['wildcard'], [(0, 1)])

    node_indices = list(range(num_nodes))
    super_node_indices = [super_node_index] * num_nodes

    data['edge']['source'] = np.append(
        data['edge']['source'], node_indices + super_node_indices
    )
    data['edge']['target'] = np.append(
        data['edge']['target'], super_node_indices + node_indices
    )
    
    total_num_edges = data['edge']['source'].shape[0]
    num_super_edges = (total_num_edges - num_edges) 
    data['edge']['super'] = np.asarray(
        [False] * num_edges + [True] * num_super_edges
    )

    if 'feature' in data['edge']:
        data['edge']['feature'] = np.pad(
            data['edge']['feature'], [(0, num_super_edges), (0, 0)]
        )
        
    return data

def _convert_dtypes(data: dict[str, dict[str, np.ndarray]]) -> np.ndarray:
    for outer_key, inner_dict in data.items():
        for inner_key, inner_value in inner_dict.items():
            if inner_key in ['source', 'target', 'size']:
                data[outer_key][inner_key] = inner_value.astype(np.int32)
            elif np.issubdtype(inner_value.dtype, np.integer):
                data[outer_key][inner_key] = inner_value.astype(np.int32)
            elif np.issubdtype(inner_value.dtype, np.floating):
                data[outer_key][inner_key] = inner_value.astype(np.float32)
    return data

def _unpack_inputs(inputs) -> tuple:
    if isinstance(inputs, np.ndarray):
        inputs = tuple(inputs.tolist())
    elif isinstance(inputs, list):
        inputs = tuple(inputs)
    if not isinstance(inputs, tuple):
        mol, context = inputs, {}
    elif isinstance(inputs[0], int) and isinstance(inputs[1], pd.Series):
        index, series = inputs
        mol = series.values[0]
        context = dict(
            zip(
                map(_snake_case, series.index[1:]), 
                map(np.asarray, series.values[1:])
            )
        )
        context['index'] = np.asarray(index)
    else:
        mol, *context = inputs
        context = dict(zip(['label', 'sample_weight'], map(np.asarray, context)))
    return mol, context
    
def _snake_case(x: str) -> str:
    return '_'.join(x.lower().split())

def _call_kwargs(func) -> bool:
    signature = inspect.signature(func)
    return any(
        (param.kind == inspect.Parameter.VAR_KEYWORD) or (param.name == 'context')
        for param in signature.parameters.values()
    )


import warnings
import keras
import numpy as np
import tensorflow as tf
from keras import backend


@keras.saving.register_keras_serializable(package='molcraft')
def gather(
    node_feature: tf.Tensor, 
    edge: tf.Tensor
) -> tf.Tensor:
    if backend.backend() == 'tensorflow':
        return tf.gather(node_feature, edge)
    expected_rank = len(keras.ops.shape(node_feature))
    current_rank = len(keras.ops.shape(edge))
    for _ in range(expected_rank - current_rank):
        edge = keras.ops.expand_dims(edge, axis=-1)
    return keras.ops.take_along_axis(node_feature, edge, axis=0)

@keras.saving.register_keras_serializable(package='molcraft')
def aggregate(
    node_feature: tf.Tensor, 
    edge: tf.Tensor, 
    num_nodes: tf.Tensor,
    mode: str = 'sum',
) -> tf.Tensor:
    if mode == 'mean':
        return segment_mean(
            node_feature, edge, num_nodes, sorted=False
        )
    return keras.ops.segment_sum(
        node_feature, edge, num_nodes, sorted=False
    )

@keras.saving.register_keras_serializable(package='molcraft')
def propagate(
    node_feature: tf.Tensor,
    edge_source: tf.Tensor,
    edge_target: tf.Tensor,
    edge_feature: tf.Tensor | None = None,
    edge_weight: tf.Tensor | None = None,
) -> tf.Tensor:
    num_nodes = keras.ops.shape(node_feature)[0]

    node_feature_source = gather(node_feature, edge_source)
    
    if edge_weight is not None:
        node_feature_source *= edge_weight

    if edge_feature is not None:
        node_feature_source += edge_feature
        
    return aggregate(node_feature, edge_target, num_nodes)

@keras.saving.register_keras_serializable(package='molcraft')
def scatter_update(
    inputs: tf.Tensor,
    indices: tf.Tensor,
    updates: tf.Tensor,
) -> tf.Tensor:
    if indices.dtype == tf.bool:
        indices = keras.ops.stack(keras.ops.where(indices), axis=-1)
    expected_rank = len(keras.ops.shape(inputs))
    current_rank = len(keras.ops.shape(indices))
    for _ in range(expected_rank - current_rank):
        indices = keras.ops.expand_dims(indices, axis=-1)
    return keras.ops.scatter_update(inputs, indices, updates)

@keras.saving.register_keras_serializable(package='molcraft')
def scatter_add(
    inputs: tf.Tensor,
    indices: tf.Tensor,
    updates: tf.Tensor,
) -> tf.Tensor:
    if indices.dtype == tf.bool:
        indices = keras.ops.stack(keras.ops.where(indices), axis=-1)
    expected_rank = len(keras.ops.shape(inputs))
    current_rank = len(keras.ops.shape(indices))
    for _ in range(expected_rank - current_rank):
        indices = keras.ops.expand_dims(indices, axis=-1)
    if backend.backend() == 'tensorflow':
        return tf.tensor_scatter_nd_add(inputs, indices, updates)
    updates = scatter_update(keras.ops.zeros_like(inputs), indices, updates)
    return inputs + updates

@keras.saving.register_keras_serializable(package='molcraft')
def edge_softmax(
    score: tf.Tensor, 
    edge_target: tf.Tensor
) -> tf.Tensor:
    num_segments = keras.ops.cond(
        keras.ops.greater(keras.ops.shape(edge_target)[0], 0),
        lambda: keras.ops.maximum(keras.ops.max(edge_target) + 1, 1),
        lambda: 0
    )
    score_max = keras.ops.segment_max(
        score, edge_target, num_segments, sorted=False
    )
    score_max = gather(score_max, edge_target)
    numerator = keras.ops.exp(score - score_max)
    denominator = keras.ops.segment_sum(
        numerator, edge_target, num_segments, sorted=False
    )
    denominator = gather(denominator, edge_target)
    return numerator / denominator

@keras.saving.register_keras_serializable(package='molcraft')
def edge_weight(
    edge: tf.Tensor,
    edge_weight: tf.Tensor,
) -> tf.Tensor:
    expected_rank = len(keras.ops.shape(edge))
    current_rank = len(keras.ops.shape(edge_weight))
    for _ in range(expected_rank - current_rank):
        edge_weight = keras.ops.expand_dims(edge_weight, axis=-1)
    return edge * edge_weight

@keras.saving.register_keras_serializable(package='molcraft')
def segment_mean(
    data: tf.Tensor,
    segment_ids: tf.Tensor,
    num_segments: int | None = None,
    sorted: bool = False,
) -> tf.Tensor:
    if num_segments is None:
        num_segments = keras.ops.cond(
            keras.ops.greater(keras.ops.shape(segment_ids)[0], 0),
            lambda: keras.ops.max(segment_ids) + 1,
            lambda: 0
        )
    if backend.backend() == 'tensorflow':
        segment_mean_fn = (
            tf.math.unsorted_segment_mean if not sorted else
            tf.math.segment_mean
        )
        return segment_mean_fn(
            data=data,
            segment_ids=segment_ids,
            num_segments=num_segments
        )
    x = keras.ops.segment_sum(
        data=data, 
        segment_ids=segment_ids, 
        num_segments=num_segments,
        sorted=sorted
    )
    sizes = keras.ops.cast(
        keras.ops.bincount(segment_ids, minlength=num_segments), 
        dtype=x.dtype
    )
    return x / sizes[:, None]

@keras.saving.register_keras_serializable(package='molcraft')
def gaussian(
    x: tf.Tensor, 
    mean: tf.Tensor, 
    std: tf.Tensor
) -> tf.Tensor:
    expected_rank = len(keras.ops.shape(x))
    current_rank = len(keras.ops.shape(mean))
    for _ in range(expected_rank - current_rank):
        mean = keras.ops.expand_dims(mean, axis=0)
        std = keras.ops.expand_dims(std, axis=0)
    a = (2 * np.pi) ** 0.5
    return keras.ops.exp(-0.5 * (((x - mean) / std) ** 2)) / (a * std)

@keras.saving.register_keras_serializable(package='molcraft')
def euclidean_distance(
    x1: tf.Tensor, 
    x2: tf.Tensor, 
    axis: int = -1,
    keepdims: bool = True,
) -> tf.Tensor:
    relative_distance = keras.ops.subtract(x1, x2)
    return keras.ops.sqrt(
        keras.ops.sum(
            keras.ops.square(relative_distance), 
            axis=axis, 
            keepdims=keepdims
        )
    )

@keras.saving.register_keras_serializable(package='molcraft')
def displacement(
    x1: tf.Tensor,
    x2: tf.Tensor,
    normalize: bool = False,
    axis: int = -1,
    keepdims: bool = True,
) -> tf.Tensor:
    displacement = keras.ops.subtract(x1, x2)
    if not normalize:
        return displacement
    return displacement / euclidean_distance(x1, x2, axis=axis, keepdims=keepdims)



__version__ = '0.1.0rc9'

import os
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

from molcraft import chem
from molcraft import features
from molcraft import descriptors
from molcraft import featurizers
from molcraft import layers 
from molcraft import models 
from molcraft import ops 
from molcraft import records 
from molcraft import tensors
from molcraft import callbacks
from molcraft import datasets
from molcraft import losses
from molcraft import trainers


import unittest 
import keras

from molcraft import losses


class TestLoss(unittest.TestCase):

    def test_gaussian_nll(self):
        loss = losses.GaussianNegativeLogLikelihood(events=3)
        value = loss(
            keras.ops.array([[1., 2., 3.]]),
            keras.ops.array([[2., 3., 4., 0.1, 0.2, 0.3]])
        )
        self.assertGreater(value, 0)
        self.assertEqual(len(keras.ops.shape(value)), 0)

        loss = losses.GaussianNegativeLogLikelihood(events=1)
        value = loss(
            keras.ops.array([1., 2., 3.]),
            keras.ops.array([[2., 0.1], [4., 0.2], [5., 0.3]])
        )
        self.assertGreater(value, 0)
        self.assertEqual(len(keras.ops.shape(value)), 0)


if __name__ == '__main__':
    unittest.main()


import unittest 
import tempfile
import shutil

from molcraft import features
from molcraft import descriptors
from molcraft import featurizers


class TestFeaturizer(unittest.TestCase):

    def setUp(self):

        self._smiles_no_atom = [''] 
        self._smiles_single_atom = ['C']
        self._smiles_single_hs_atom = ['[H]']
        self._smiles_two_disconnected_singles = ['C.O']
        self._smiles_two_disconnected_doubles = ['CC.CO']
        self._smiles_one_molecule = [
            'C(C(=O)O)N'
        ]
        self._smiles_two_molecules = [
            'O=C([C@H](CC1=CNC=N1)N)O',
            'C(C(=O)O)N'
        ]
        self._smiles_single_double = [
            'C',
            'CO'
        ]

    def test_mol_featurizer(self):
            
        featurizer = featurizers.MolGraphFeaturizer(
            atom_features=[
                features.AtomType({'C', 'N', 'O', 'H'}),
                features.NumHydrogens({0, 1, 2, 3, 4})
            ],
            bond_features=[
                features.BondType({'single', 'double', 'aromatic'}),
                features.IsRotatable(),
            ],
            molecule_features='auto',
            super_node=True,
            self_loops=False,
            include_hydrogens=False, 
        ) 

        tmp_dir = tempfile.mkdtemp()
        tmp_file = tmp_dir + '/featurizer.json'
        featurizers.save_featurizer(featurizer, tmp_file)
        _ = featurizers.load_featurizer(tmp_file)
        shutil.rmtree(tmp_dir)
        
        node_dim = 9
        edge_dim = 4

        smiles = self._smiles_one_molecule
        num_nodes = (5 + 1)
        num_edges = (8 + 5 * 2)
        with self.subTest(smiles=smiles):
            graph = featurizer(smiles)
            self.assertEqual(graph.context['feature'].shape, (1, 10))
            self.assertEqual(graph.node['feature'].shape, (num_nodes, node_dim))
            self.assertEqual(graph.edge['feature'].shape, (num_edges, edge_dim))
            self.assertEqual(graph.node['feature'].dtype.name, 'float32')
            self.assertEqual(graph.edge['feature'].dtype.name, 'float32')
            self.assertEqual(graph.edge['source'].dtype.name, 'int32')
            self.assertEqual(graph.edge['target'].dtype.name, 'int32')

        smiles = self._smiles_two_molecules
        num_nodes = (5 + 1) + (11 + 1)
        num_edges = (8 + 5 * 2) + (22 + 11 * 2)
        with self.subTest(smiles=smiles):
            graph = featurizer(smiles)
            self.assertEqual(graph.context['feature'].shape, (2, 10))
            self.assertEqual(graph.node['feature'].shape, (num_nodes, node_dim))
            self.assertEqual(graph.edge['feature'].shape, (num_edges, edge_dim))
            self.assertEqual(graph.node['feature'].dtype.name, 'float32')
            self.assertEqual(graph.edge['feature'].dtype.name, 'float32')
            self.assertEqual(graph.edge['source'].dtype.name, 'int32')
            self.assertEqual(graph.edge['target'].dtype.name, 'int32')
        
        smiles = self._smiles_no_atom 
        num_nodes = (0 + 1)
        num_edges = (0 + 0 * 2)
        with self.subTest(smiles=smiles):
            graph = featurizer(smiles)
            self.assertEqual(graph.context['feature'].shape, (1, 10))
            self.assertEqual(graph.node['feature'].shape, (num_nodes, node_dim))
            self.assertEqual(graph.edge['feature'].shape, (num_edges, edge_dim))
            self.assertEqual(graph.node['feature'].dtype.name, 'float32')
            self.assertEqual(graph.edge['feature'].dtype.name, 'float32')
            self.assertEqual(graph.edge['source'].dtype.name, 'int32')
            self.assertEqual(graph.edge['target'].dtype.name, 'int32')

        smiles = self._smiles_single_atom 
        num_nodes = (1 + 1)
        num_edges = (0 + 1 * 2)
        with self.subTest(smiles=smiles):
            graph = featurizer(smiles)
            self.assertEqual(graph.context['feature'].shape, (1, 10))
            self.assertEqual(graph.node['feature'].shape, (num_nodes, node_dim))
            self.assertEqual(graph.edge['feature'].shape, (num_edges, edge_dim))
            self.assertEqual(graph.node['feature'].dtype.name, 'float32')
            self.assertEqual(graph.edge['feature'].dtype.name, 'float32')
            self.assertEqual(graph.edge['source'].dtype.name, 'int32')
            self.assertEqual(graph.edge['target'].dtype.name, 'int32')

        smiles = self._smiles_single_hs_atom 
        num_nodes = (1 + 1)
        num_edges = (0 + 1 * 2)
        with self.subTest(smiles=smiles):
            graph = featurizer(smiles)
            self.assertEqual(graph.context['feature'].shape, (1, 10))
            self.assertEqual(graph.node['feature'].shape, (num_nodes, node_dim))
            self.assertEqual(graph.edge['feature'].shape, (num_edges, edge_dim))
            self.assertEqual(graph.node['feature'].dtype.name, 'float32')
            self.assertEqual(graph.edge['feature'].dtype.name, 'float32')
            self.assertEqual(graph.edge['source'].dtype.name, 'int32')
            self.assertEqual(graph.edge['target'].dtype.name, 'int32')

    def test_mol_featurizer3d(self):
        
        num_conformers = 1
        featurizer = featurizers.MolGraphFeaturizer3D(
            atom_features=[
                features.AtomType({'C', 'N', 'O', 'H'}, encode_oov=True),
                features.NumHydrogens({0, 1, 2, 3, 4})
            ],
            pair_features=[
                features.PairDistance(max_distance=20)
            ],
            molecule_features=[
                descriptors.ForceFieldEnergy(),
            ],
            super_node=True,
            self_loops=False,
            include_hydrogens=False, 
            radius=5.0, 
        ) 

        tmp_dir = tempfile.mkdtemp()
        tmp_file = tmp_dir + '/featurizer.json'
        featurizers.save_featurizer(featurizer, tmp_file)
        _ = featurizers.load_featurizer(tmp_file)
        shutil.rmtree(tmp_dir)

        node_dim = 10
        edge_dim = 22

        smiles = self._smiles_one_molecule
        num_nodes = (5 + 1) * num_conformers
        num_edges = (8 + 5 * 2) * num_conformers
        with self.subTest(smiles=smiles):
            graph = featurizer(smiles)
            self.assertEqual(graph.node['feature'].shape, (num_nodes, node_dim))
            self.assertEqual(graph.node['coordinate'].shape, (num_nodes, 3))
            self.assertGreaterEqual(graph.edge['feature'].shape[0], num_edges)
            self.assertEqual(graph.edge['feature'].shape[-1], edge_dim)
            self.assertEqual(graph.node['feature'].dtype.name, 'float32')
            self.assertEqual(graph.edge['feature'].dtype.name, 'float32')
            self.assertEqual(graph.edge['source'].dtype.name, 'int32')
            self.assertEqual(graph.edge['target'].dtype.name, 'int32')

        smiles = self._smiles_two_molecules
        num_nodes = ((5 + 1) + (11 + 1)) * num_conformers
        num_edges = ((8 + 5 * 2) + (22 + 11 * 2)) * num_conformers
        with self.subTest(smiles=smiles):
            graph = featurizer(smiles)
            self.assertEqual(graph.node['feature'].shape, (num_nodes, node_dim))
            self.assertEqual(graph.node['coordinate'].shape, (num_nodes, 3))
            self.assertGreaterEqual(graph.edge['feature'].shape[0], num_edges)
            self.assertEqual(graph.edge['feature'].shape[-1], edge_dim)
            self.assertEqual(graph.node['feature'].dtype.name, 'float32')
            self.assertEqual(graph.edge['feature'].dtype.name, 'float32')
            self.assertEqual(graph.edge['source'].dtype.name, 'int32')
            self.assertEqual(graph.edge['target'].dtype.name, 'int32')

        smiles = self._smiles_single_atom 
        num_nodes = (1 + 1) * num_conformers
        num_edges = (0 + 1 * 2) * num_conformers
        with self.subTest(smiles=smiles):
            graph = featurizer(smiles)
            self.assertEqual(graph.node['feature'].shape, (num_nodes, node_dim))
            self.assertEqual(graph.edge['feature'].shape, (num_edges, edge_dim))
            self.assertEqual(graph.node['coordinate'].shape, (num_nodes, 3))
            self.assertEqual(graph.node['feature'].dtype.name, 'float32')
            self.assertEqual(graph.edge['feature'].dtype.name, 'float32')
            self.assertEqual(graph.edge['source'].dtype.name, 'int32')
            self.assertEqual(graph.edge['target'].dtype.name, 'int32')

        smiles = self._smiles_single_hs_atom 
        num_nodes = (1 + 1) * num_conformers
        num_edges = (0 + 1 * 2) * num_conformers
        with self.subTest(smiles=smiles):
            graph = featurizer(smiles)
            self.assertEqual(graph.node['feature'].shape, (num_nodes, node_dim))
            self.assertEqual(graph.edge['feature'].shape, (num_edges, edge_dim))
            self.assertEqual(graph.node['coordinate'].shape, (num_nodes, 3))
            self.assertEqual(graph.node['feature'].dtype.name, 'float32')
            self.assertEqual(graph.edge['feature'].dtype.name, 'float32')
            self.assertEqual(graph.edge['source'].dtype.name, 'int32')
            self.assertEqual(graph.edge['target'].dtype.name, 'int32')


if __name__ == '__main__':
    unittest.main()


import warnings
import os
import math
import glob
import time
import typing 
import tensorflow as tf
import numpy as np
import pandas as pd
import multiprocessing as mp

from molcraft import tensors

if typing.TYPE_CHECKING:
    from molcraft import featurizers


def write(
    inputs: list[str | tuple], 
    featurizer: 'featurizers.GraphFeaturizer',
    path: str, 
    exist_ok: bool = False,
    overwrite: bool = False, 
    num_files: typing.Optional[int] = None, 
    num_processes: typing.Optional[int] = None,
    multiprocessing: bool = False,
    device: str = '/cpu:0'
) -> None:
    
    if os.path.isdir(path):
        if not exist_ok:
            raise FileExistsError(f'Records already exist: {path}')
        if not overwrite:
            return
        else:
            _remove_files(path)
    else:
        os.makedirs(path)

    with tf.device(device):
        
        if isinstance(inputs, (pd.DataFrame, pd.Series)):
            inputs = list(inputs.iterrows())

        example = featurizer._call(inputs[0])
        save_spec(os.path.join(path, 'spec.pb'), example.spec)

        if num_processes is None:
            num_processes = mp.cpu_count()

        if num_files is None:
            num_files = min(len(inputs), max(1, math.ceil(len(inputs) / 1_000)))
            
        num_examples = len(inputs)
        chunk_sizes = [0] * num_files
        for i in range(num_examples):
            chunk_sizes[i % num_files] += 1
        
        input_chunks = []
        start_indices = []
        current_index = 0
        for size in chunk_sizes:
            input_chunks.append(inputs[current_index: current_index + size])
            start_indices.append(current_index)
            current_index += size 
        
        assert current_index == num_examples
        
        paths = [
            os.path.join(path, f'tfrecord-{i:06d}.tfrecord')
            for i in range(num_files)
        ]
        
        if not multiprocessing:
            for path, input_chunk, start_index in zip(paths, input_chunks, start_indices):
                _write_tfrecord(input_chunk, path, featurizer, start_index)
            return
        
        processes = []
        
        for path, input_chunk, start_index in zip(paths, input_chunks, start_indices):
        
            while len(processes) >= num_processes:
                for process in processes:
                    if not process.is_alive():
                        processes.remove(process)
                else:
                    time.sleep(0.1)
                    continue
                    
            process = mp.Process(
                target=_write_tfrecord,
                args=(input_chunk, path, featurizer, start_index)
            )
            processes.append(process)
            process.start()

        for process in processes:
            process.join()         
    
def read(
    path: str, 
    shuffle_files: bool = False
) -> tf.data.Dataset:
    spec = load_spec(os.path.join(path, 'spec.pb'))
    filenames = sorted(glob.glob(os.path.join(path, '*.tfrecord')))
    num_files = len(filenames)
    ds = tf.data.Dataset.from_tensor_slices(filenames)
    if shuffle_files:
        ds = ds.shuffle(num_files)
    ds = ds.interleave(
        tf.data.TFRecordDataset, num_parallel_calls=1)
    ds = ds.map(
        lambda x: _parse_example(x, spec),
        num_parallel_calls=tf.data.AUTOTUNE)
    if not tensors.is_scalar(spec):
        ds = ds.unbatch()
    return ds

def save_spec(path: str, spec: tensors.GraphTensor.Spec) -> None:
    proto = spec.experimental_as_proto()
    with open(path, 'wb') as fh:
        fh.write(proto.SerializeToString())

def load_spec(path: str) -> tensors.GraphTensor.Spec:
    with open(path, 'rb') as fh:
        serialized_proto = fh.read()
    spec = tensors.GraphTensor.Spec.experimental_from_proto(
        tensors.GraphTensor.Spec
        .experimental_type_proto()
        .FromString(serialized_proto)
    )
    return spec
    
def _write_tfrecord(
    inputs: list, 
    path: str,
    featurizer: 'featurizers.GraphFeaturizer', 
    start_index: int,
) -> None:
    with tf.io.TFRecordWriter(path) as writer:
        for i, x in enumerate(inputs):
            try:
                tensor = featurizer._call(x)
                serialized = _serialize_example(tensor)
                writer.write(serialized)
            except Exception as e:
                index = getattr(x, 'Index', (i + start_index))
                warnings.warn(
                    f'Could not write record for index {index}, '
                    f'proceeding without it. Exception raised:\n{e}'
                )

def _serialize_example(tensor):
    flat_values = tf.nest.flatten(tensor, expand_composites=True)
    flat_values = [
        tf.io.serialize_tensor(value).numpy() for value in flat_values
    ]
    feature = tf.train.Feature(
        bytes_list=tf.train.BytesList(value=flat_values)
    )
    example_proto = tf.train.Example(
        features=tf.train.Features(feature={'feature': feature})
    )
    return example_proto.SerializeToString()
    
def _parse_example(
    x: tf.Tensor, 
    spec: tensors.GraphTensor.Spec
) -> tf.Tensor:
    out = tf.io.parse_single_example(
        x, features={'feature': tf.io.RaggedFeature(tf.string)})['feature']
    out = [
        tf.ensure_shape(tf.io.parse_tensor(x[0], s.dtype), s.shape) 
        for (x, s) in zip(
            tf.split(out, len(tf.nest.flatten(spec, expand_composites=True))), 
            tf.nest.flatten(spec, expand_composites=True)
        )
    ]
    out = tf.nest.pack_sequence_as(spec, tf.nest.flatten(out), expand_composites=True)
    return out

def _remove_files(path):
    for filename in os.listdir(path):
        if filename.endswith('tfrecord') or filename == 'spec.pb':
            filepath = os.path.join(path, filename)
            os.remove(filepath)



import warnings
import typing
import keras
import numpy as np
import tensorflow as tf
from pathlib import Path
from keras.src.models import functional

from molcraft import layers
from molcraft import tensors
from molcraft import ops


@keras.saving.register_keras_serializable(package="molcraft")
class GraphModel(layers.GraphLayer, keras.models.Model):
    
    """A graph model.

    Currently, the `GraphModel` only supports `GraphTensor` input.

    Build a subclassed GraphModel:

    >>> import molcraft
    >>> import keras
    >>>
    >>> featurizer = molcraft.featurizers.MolGraphFeaturizer()
    >>> graph = featurizer([('N[C@@H](C)C(=O)O', 1.0), ('N[C@@H](CS)C(=O)O', 2.0)])
    >>>
    >>> @keras.saving.register_keras_serializable()
    >>> class GraphNeuralNetwork(molcraft.models.GraphModel):
    ...     def __init__(self, units, **kwargs):
    ...         super().__init__(**kwargs)
    ...         self.units = units
    ...         self.node_embedding = molcraft.layers.NodeEmbedding(self.units)
    ...         self.edge_embedding = molcraft.layers.EdgeEmbedding(self.units)
    ...         self.conv_1 = molcraft.layers.GraphConv(self.units)
    ...         self.conv_2 = molcraft.layers.GraphConv(self.units)
    ...         self.readout = molcraft.layers.Readout('mean')
    ...         self.dense = keras.layers.Dense(1)
    ...     def propagate(self, graph):
    ...         x = self.edge_embedding(self.node_embedding(graph))
    ...         x = self.conv_2(self.conv_1(x))
    ...         return self.dense(self.readout(x))
    ...     def get_config(self):
    ...         config = super().get_config()
    ...         config['units'] = self.units
    ...         return config
    >>>
    >>> model = GraphNeuralNetwork(128)
    >>> model.compile(
    ...     optimizer=keras.optimizers.Adam(1e-3),
    ...     loss=keras.losses.MeanSquaredError(),
    ...     metrics=[keras.metrics.MeanAbsolutePercentageError(name='mape')]
    ... )
    >>> model.fit(graph, epochs=10)
    >>> mse, mape = model.evaluate(graph)
    >>> preds = model.predict(graph)

    Build a functional GraphModel:

    >>> import molcraft
    >>> import keras
    >>>
    >>> featurizer = molcraft.featurizers.MolGraphFeaturizer()
    >>> graph = featurizer([('N[C@@H](C)C(=O)O', 1.0), ('N[C@@H](CS)C(=O)O', 2.0)])
    >>>
    >>> inputs = molcraft.layers.Input(graph.spec)
    >>> x = molcraft.layers.NodeEmbedding(128)(inputs)
    >>> x = molcraft.layers.EdgeEmbedding(128)(x)
    >>> x = molcraft.layers.GraphConv(128)(x)
    >>> x = molcraft.layers.GraphConv(128)(x)
    >>> x = molcraft.layers.Readout('mean')(x)
    >>> outputs = keras.layers.Dense(1)(x)
    >>> model = molcraft.models.GraphModel(inputs, outputs)
    >>> model.compile(
    ...     optimizer=keras.optimizers.Adam(1e-3),
    ...     loss=keras.losses.MeanSquaredError(),
    ...     metrics=[keras.metrics.MeanAbsolutePercentageError(name='mape')]
    ... )
    >>> model.fit(graph, epochs=10)
    >>> mse, mape = model.evaluate(graph)
    >>> preds = model.predict(graph)

    Build a GraphModel using `from_layers`:

    >>> import molcraft 
    >>> import keras
    >>> 
    >>> featurizer = molcraft.featurizers.MolGraphFeaturizer()
    >>> graph = featurizer([('N[C@@H](C)C(=O)O', 1.0), ('N[C@@H](CS)C(=O)O', 2.0)])
    >>> 
    >>> model = molcraft.models.GraphModel.from_layers([
    ...     molcraft.layers.Input(graph.spec),
    ...     molcraft.layers.NodeEmbedding(128),
    ...     molcraft.layers.EdgeEmbedding(128),
    ...     molcraft.layers.GraphConv(128),
    ...     molcraft.layers.GraphConv(128),
    ...     molcraft.layers.Readout('mean'),
    ...     keras.layers.Dense(1)
    ... ])
    >>> model.compile(
    ...     optimizer=keras.optimizers.Adam(1e-3),
    ...     loss=keras.losses.MeanSquaredError(),
    ...     metrics=[keras.metrics.MeanAbsolutePercentageError(name='mape')]
    ... )
    >>> model.fit(graph, epochs=10)
    >>> mse, mape = model.evaluate(graph)
    >>> preds = model.predict(graph)

    """

    def __new__(cls, *args, **kwargs):
        if _functional_init_arguments(args, kwargs) and cls == GraphModel:
            return FunctionalGraphModel(*args, **kwargs)
        return super().__new__(cls)
    
    def __init__(self, *args, **kwargs):
        self._model_layers = kwargs.pop('model_layers', None)
        super().__init__(*args, **kwargs)
        self.jit_compile = False 

    @classmethod
    def from_layers(cls, graph_layers: list, **kwargs):
        """Creates a graph model from a list of graph layers.

        Currently requires `molcraft.layers.Input(spec)`. 

        If `molcraft.layers.Input(spec)` is supplied, it both 
        creates and builds the layer, as a functional model.
        `molcraft.layers.Input` is a function which returns
        a nested structure of graph components based on `spec`.

        Args:
            graph_layers:
                A list of `GraphLayer` instances, except the initial element
                which is a dictionary of Keras tensors produced by 
                `molcraft.layers.Input(spec)`.
        """
        if not tensors.is_graph(graph_layers[0]):
            return cls(model_layers=graph_layers, **kwargs)
        elif cls != GraphModel:
            return cls(model_layers=graph_layers[1:], **kwargs)
        inputs: dict = graph_layers.pop(0)
        x = inputs
        for layer in graph_layers:
            if isinstance(layer, list):
                layer = layers.GraphNetwork(layer)
            x = layer(x)
        outputs = x
        return cls(inputs=inputs, outputs=outputs, **kwargs)

    def propagate(self, graph: tensors.GraphTensor) -> tensors.GraphTensor:
        if self._model_layers is None:
            return super().propagate(graph)
        for layer in self._model_layers:
            graph = layer(graph)
        return graph
    
    def get_config(self):
        """Obtain model config."""
        config = super().get_config()
        if hasattr(self, '_model_layers') and self._model_layers is not None:
            config['model_layers'] = [
                keras.saving.serialize_keras_object(l) 
                for l in self._model_layers
            ]
        return config 
    
    @classmethod
    def from_config(cls, config: dict):
        """Obtain model from model config."""
        if 'model_layers' in config:
            config['model_layers'] = [
                keras.saving.deserialize_keras_object(l) 
                for l in config['model_layers']
            ]
        return super().from_config(config)

    def compile(
        self, 
        optimizer: keras.optimizers.Optimizer | str | None = 'rmsprop',
        loss: keras.losses.Loss | str | None = None,
        loss_weights: dict[str, float] = None,
        metrics: list[keras.metrics.Metric] = None,
        weighted_metrics: list[keras.metrics.Metric] | None = None,
        run_eagerly: bool = False,
        steps_per_execution: int = 1,
        jit_compile: str | bool = False,
        auto_scale_loss: bool = True,
        **kwargs
    ) -> None:
        """Compiles the model.

        Args:
            optimizer:
                The optimizer to be used (a `keras.optimizers.Optimizer` subclass).
            loss:
                The loss function to be used (a `keras.losses.Loss` subclass).
            metrics:
                A list of metrics to be used during training (`fit`) and evaluation
                (`evaluate`). Should be `keras.metrics.Metric` subclasses.
            kwargs:
                See `Model.compile` in Keras documentation. 
                May or may not apply here.
        """
        super().compile(
            optimizer=optimizer,
            loss=loss,
            loss_weights=loss_weights,
            metrics=metrics,
            weighted_metrics=weighted_metrics,
            run_eagerly=run_eagerly,
            steps_per_execution=steps_per_execution,
            jit_compile=jit_compile,
            auto_scale_loss=auto_scale_loss,
            **kwargs
        )

    def fit(self, x: tensors.GraphTensor | tf.data.Dataset, **kwargs):
        """Fits the model.

        Args:
            x: 
                A `GraphTensor` instance or a `tf.data.Dataset` constructed from
                a `GraphTensor` instance. In comparison to a typical Keras model,
                the label (typically denoted `y`) and the sample_weight (typically
                denoted `sample_weight`) should be encoded in the context of the 
                `GraphTensor` instance, as `label` and `weight` respectively.
            validation_data:
                A `GraphTensor` instance or a `tf.data.Dataset` constructed from
                a `GraphTensor` instance. In comparison to a typical Keras model,
                the label (typically denoted `y`) and the sample_weight (typically
                denoted `sample_weight`) should be encoded in the context of the 
                `GraphTensor` instance, as `label` and `weight` respectively.
            validaton_split:
                The fraction of training data to be used as validation data. 
                Only works if a `GraphTensor` instance is passed as `x`. 
            batch_size:
                Number of samples per batch of computation.
            epochs:
                Number of iterations over the entire dataset.
            callbacks:
                A list of callbacks to apply during training.
            kwargs:
                See `Model.fit` in Keras documentation. 
                May or may not apply here. 
        """
        batch_size = kwargs.get('batch_size', 32)
        x_val = kwargs.pop('validation_data', None)
        val_split = kwargs.pop('validation_split', None)
        if x_val is not None and isinstance(x_val, tensors.GraphTensor):
            x_val = _make_dataset(x_val, batch_size)
        if isinstance(x, tensors.GraphTensor):
            if val_split:
                val_size = int(val_split * x.num_subgraphs)
                x_val = _make_dataset(x[-val_size:], batch_size)
                x = x[:-val_size]
            x = _make_dataset(x, batch_size, shuffle=kwargs.get('shuffle', True))
        return super().fit(x, validation_data=x_val, **kwargs)
    
    def evaluate(self, x: tensors.GraphTensor | tf.data.Dataset, **kwargs):
        """Evaluation of the model.

        Args:
            x: 
                A `GraphTensor` instance or a `tf.data.Dataset` constructed from
                a `GraphTensor` instance. In comparison to a typical Keras model,
                the label (typically denoted `y`) and the sample_weight (typically
                denoted `sample_weight`) should be encoded in the context of the 
                `GraphTensor` instance, as `label` and `weight` respectively.
            batch_size:
                Number of samples per batch of computation.
            kwargs:
                See `Model.evaluate` in Keras documentation. 
                May or may not apply here. 
        """
        batch_size = kwargs.get('batch_size', 32)
        if isinstance(x, tensors.GraphTensor):
            x = _make_dataset(x, batch_size)
        metric_results = super().evaluate(x, **kwargs)
        return tf.nest.map_structure(lambda value: float(value), metric_results)
    
    def predict(self, x: tensors.GraphTensor | tf.data.Dataset, **kwargs):
        """Makes predictions with the model.

        Args:
            x: 
                A `GraphTensor` instance or a `tf.data.Dataset` constructed from
                a `GraphTensor` instance.
            batch_size:
                Number of samples per batch of computation.
            kwargs:
                See `Model.predict` in Keras documentation. 
                May or may not apply here. 
        """
        batch_size = kwargs.get('batch_size', 32)
        if isinstance(x, tensors.GraphTensor):
            x = _make_dataset(x, batch_size)
        output = super().predict(x, **kwargs)
        if tensors.is_graph(output):
            return tensors.from_dict(output).flatten()
        return output

    def get_compile_config(self) -> dict | None:
        config = super().get_compile_config()
        if config is None:
            return
        return config

    def compile_from_config(self, config: dict | None) -> None:
        if config is None:
            return
        config = keras.utils.deserialize_keras_object(config)
        self.compile(**config)
        if hasattr(self, 'optimizer') and self.built:
            self.optimizer.build(self.trainable_variables)

    def save(
        self,
        filepath: str | Path,
        *args,
        **kwargs
    ) -> None:
        """Saves an entire model.
        
        Args:
            filepath:
                A string with the path to the model file (requires `.keras` suffix)
        """
        if not self.built:
            raise ValueError('Cannot save model as it has not been built yet.')
        super().save(filepath, *args, **kwargs)

    @staticmethod
    def load(
        filepath: str | Path,
        *args,
        **kwargs
    ) -> keras.Model:
        """A `staticmethod` loading an entire model.

        Args:
            filepath:
                A string with the path to the model file (requires `.keras` suffix)
        """
        return keras.models.load_model(filepath, *args, **kwargs)
    
    def save_weights(self, filepath, *args, **kwargs):
        """Saves the weights of the model.

        Args:
            filepath:
                A string with the path to the file (requires `.weights.h5` suffix)
        """
        path = Path(filepath).parent
        path.mkdir(parents=True, exist_ok=True)
        return super().save_weights(filepath, *args, **kwargs)

    def load_weights(self, filepath, *args, **kwargs):
        """Loads the weights from file saved via `save_weights()`.
        
        Args:
            filepath:
                A string with the path to the file (requires `.weights.h5` suffix)
        """
        super().load_weights(filepath, *args, **kwargs)

    def embedding(self, layer_name: str = None) -> 'FunctionalGraphModel':
        model = self
        if not isinstance(model, FunctionalGraphModel):
            raise ValueError(
                'Currently, to extract the embedding part of the model, '
                'it needs to be a `FunctionalGraphModel`. '
            )
        inputs = model.input 
        if not layer_name:
            for layer in model.layers:
                if isinstance(layer, layers.Readout):
                    outputs = layer.output 
        else:
            layer = model.get_layer(layer_name)
            outputs = (
                layer.output if isinstance(layer, keras.layers.Layer) else None
            )
            if outputs is None:
                raise ValueError(
                    f'Could not find `{layer_name}` or '
                    f'`{layer_name} is not a `keras.layers.Layer`.'
                )
        return self.__class__(inputs, outputs, name=f'{self.name}_embedding')

    def backbone(self) -> 'FunctionalGraphModel':
        if not isinstance(self, FunctionalGraphModel):
            raise ValueError(
                'Currently, to extract the backbone part of the model, '
                'it needs to be a `FunctionalGraphModel`, with a `Readout` '
                'layer dividing the backbone and the head part of the model.'
            )
        inputs = self.input
        outputs = None
        for layer in self.layers:
            if isinstance(layer, layers.Readout):
                outputs = layer.output
        if outputs is None:
            raise ValueError(
                'Could not extract output. `Readout` layer not found.'
            )
        return self.__class__(inputs, outputs, name=f'{self.name}_backbone')

    def head(self) -> functional.Functional:
        if not isinstance(self, FunctionalGraphModel):
            raise ValueError(
                'Currently, to extract the head part of the model, '
                'it needs to be a `FunctionalGraphModel`, with a `Readout` '
                'layer dividing the backbone and the head part of the model.'
            )
        inputs = None
        for layer in self.layers:
            if isinstance(layer, layers.Readout):
                inputs = layer.output
        if inputs is None:
            raise ValueError(
                'Could not extract input. `Readout` layer not found.'
            )
        outputs = layer.output
        return keras.models.Model(inputs, outputs, name=f'{self.name}_head')

    def train_step(self, tensor: tensors.GraphTensor) -> dict[str, float]:
        with tf.GradientTape() as tape:
            output = self(tensor, training=True)
            y, y_pred, sample_weight = _get_loss_args(tensor, output)
            loss = self.compute_loss(tensor, y, y_pred, sample_weight)
            loss = self.optimizer.scale_loss(loss)
        trainable_weights = self.trainable_weights 
        gradients = tape.gradient(loss, trainable_weights)
        self.optimizer.apply_gradients(zip(gradients, trainable_weights))
        return self.compute_metrics(tensor, y, y_pred, sample_weight)
    
    def test_step(self, tensor: tensors.GraphTensor) -> dict[str, float]:
        output = self(tensor, training=False)
        y, y_pred, sample_weight = _get_loss_args(tensor, output)
        return self.compute_metrics(tensor, y, y_pred, sample_weight)
    
    def predict_step(self, tensor: tensors.GraphTensor) -> np.ndarray:
        output = self(tensor, training=False)
        if tensors.is_graph(output):
            if not isinstance(output, tensors.GraphTensor):
                output = tensors.from_dict(output)
            output = tensors.to_dict(output.unflatten())
        return output

    def compute_loss(self, x, y, y_pred, sample_weight=None):
        return super().compute_loss(x, y, y_pred, sample_weight)
        
    def compute_metrics(self, x, y, y_pred, sample_weight=None) -> dict[str, float]:
        loss = self.compute_loss(x, y, y_pred, sample_weight)
        metric_results = {}
        for metric in self.metrics:
            if metric.name == "loss":
                metric.update_state(loss)
                metric_results[metric.name] = metric.result()
            else:
                metric.update_state(y, y_pred, sample_weight=sample_weight)
                metric_results.update(metric.result())
        return metric_results
    

@keras.saving.register_keras_serializable(package="molcraft")
class FunctionalGraphModel(functional.Functional, GraphModel):

    @property 
    def layers(self):
        return [
            l for l in super().layers if not isinstance(l, keras.layers.InputLayer)
        ]


def save_model(model: GraphModel, filepath: str | Path, *args, **kwargs) -> None:
    if not model.built:
        raise ValueError(
            'Model and its layers have not yet been (fully) built. '
            'Build the model before saving it: `model.build(graph_spec)` '
            'or `model(graph)`.'
        )
    keras.models.save_model(model, filepath, *args, **kwargs)

def load_model(filepath: str | Path, inputs=None, *args, **kwargs) -> GraphModel:
    return keras.models.load_model(filepath, *args, **kwargs)

def create(
    *layers: list[keras.layers.Layer],
    **kwargs
) -> GraphModel:
    if isinstance(layers[0], list):
        layers = layers[0]
    return GraphModel.from_layers(
        list(layers), **kwargs
    )

def interpret(
    model: GraphModel,
    graph_tensor: tensors.GraphTensor,
) -> tensors.GraphTensor:
    x = graph_tensor
    if tensors.is_ragged(x):
        x = x.flatten()
    graph_indicator = x.graph_indicator
    y_true = x.context.get('label')
    features = []
    with tf.GradientTape(watch_accessed_variables=False) as tape:
        for layer in model.layers:
            if isinstance(layer, keras.layers.InputLayer):
                continue
            if isinstance(layer, layers.GraphNetwork):
                x, taped_features = layer.tape_propagate(x, tape, training=False)
                features.extend(taped_features)
            else:
                if (
                    isinstance(layer, layers.GraphConv) and 
                    isinstance(x, tensors.GraphTensor)
                ):
                    tape.watch(x.node['feature'])
                    features.append(x.node['feature'])
                x = layer(x, training=False)
        y_pred = x
        if y_true is not None and len(y_true.shape) > 1: 
            target = tf.gather_nd(y_pred, tf.where(y_true != 0))
        else:
            target = y_pred
    gradients = tape.gradient(target, features)
    features = keras.ops.concatenate(features, axis=-1)
    gradients = keras.ops.concatenate(gradients, axis=-1)
    alpha = ops.segment_mean(gradients, graph_indicator)
    alpha = ops.gather(alpha, graph_indicator)
    maps = keras.ops.where(gradients != 0, alpha * features, gradients)
    maps = keras.ops.sum(maps, axis=-1)
    return graph_tensor.update(
        {
            'node': {
                'saliency': maps
            }
        }
    )

def saliency(
    model: GraphModel,
    graph_tensor: tensors.GraphTensor,
) -> tensors.GraphTensor:
    x = graph_tensor
    if tensors.is_ragged(x):
        x = x.flatten()
    y_true = x.context.get('label')
    with tf.GradientTape(watch_accessed_variables=False) as tape:
        tape.watch(x.node['feature'])
        y_pred = model(x, training=False)
        if y_true is not None and len(y_true.shape) > 1: 
            target = tf.gather_nd(y_pred, tf.where(y_true != 0))
        else:
            target = y_pred
    gradients = tape.gradient(target, x.node['feature'])
    gradients = keras.ops.absolute(gradients)
    return graph_tensor.update(
        {
            'node': {
                'feature_saliency': gradients
            }
        }
    ) 
    
def _functional_init_arguments(args, kwargs):
    return (
        (len(args) == 2)
        or (len(args) == 1 and "outputs" in kwargs)
        or ("inputs" in kwargs and "outputs" in kwargs)
    )

def _make_dataset(x: tensors.GraphTensor, batch_size: int, shuffle: bool = False):
    ds = tf.data.Dataset.from_tensor_slices(x)
    if shuffle:
        ds = ds.shuffle(buffer_size=ds.cardinality())
    return ds.batch(batch_size).prefetch(-1)

def _get_loss_args(
    inputs: tensors.GraphTensor,
    outputs: tensors.GraphTensor | tf.Tensor,
) -> tuple[tf.Tensor, tf.Tensor, tf.Tensor | None]:
    if (
        not isinstance(inputs, tensors.GraphTensor) and
        tensors.is_graph(inputs)
    ):
        inputs = tensors.from_dict(inputs)
    if (
        not isinstance(outputs, tensors.GraphTensor) and
        tensors.is_graph(outputs)
    ):
        outputs = tensors.from_dict(outputs)

    if not isinstance(outputs, tensors.GraphTensor):
        tensor, prediction = inputs, outputs
    else:
        tensor, prediction = outputs, None

    if 'label' in tensor.context:
        data = tensor.context
    elif 'label' in tensor.node:
        data = tensor.node
    elif 'label' in tensor.edge:
        data = tensor.edge
    else:
        raise ValueError(
            'Could not find a `label` in the `GraphTensor`. Make sure a '
            '`label` exists in either the `context`, `node` or `edge`.'
        )

    prediction = (
        prediction if prediction is not None else data.get('prediction')
    )
    if prediction is None:
        raise ValueError(
            'Could not find a `prediction` in the `GraphTensor`. Make sure a '
            '`prediction` exists in either the `context`, `node` or `edge`.'
        )
    return data['label'], prediction, data.get('sample_weight')



import unittest 

import keras

from molcraft import tensors 
from molcraft import layers 


class TestLayer(unittest.TestCase):

    def setUp(self):

        self.tensors = [
            # Graph with two subgraphs
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([2, 3], dtype='int32')
                },
                node={
                    'feature': keras.ops.array([[1., 2.], [3., 4.], [5., 6.], [6., 7.], [8., 9.]], dtype='float32'),
                    'weight': keras.ops.array([0.50, 1.00, 2.00, 0.25, 0.75], dtype='float32')
                },
                edge={
                    'source': keras.ops.array([0, 1, 3, 4, 4, 3, 2], dtype='int32'),
                    'target': keras.ops.array([1, 0, 2, 3, 4, 4, 3], dtype='int32'),
                    'feature': keras.ops.array([[1.], [2.], [3.], [4.], [5.], [6.], [7.]], dtype='float32')
                }
            ),
            # Graph with two subgraphs, none which has edges
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([2, 3], dtype='int32')
                },
                node={
                    'feature': keras.ops.array([[1., 2.], [3., 4.], [5., 6.], [6., 7.], [8., 9.]], dtype='float32'),
                    'weight': keras.ops.array([0.50, 1.00, 2.00, 0.25, 0.75], dtype='float32')
                },
                edge={
                    'source': keras.ops.zeros([0], dtype='int32'),
                    'target': keras.ops.zeros([0], dtype='int32'),
                    'feature': keras.ops.zeros([0, 1], dtype='float32')
                }
            ),
            # Graph with two subgraphs, none of which has nodes or edges
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([0, 0], dtype='int32')
                },
                node={
                    'feature': keras.ops.zeros([0, 2], dtype='float32'),
                    'weight': keras.ops.zeros([0], dtype='float32')
                },
                edge={
                    'source': keras.ops.zeros([0], dtype='int32'),
                    'target': keras.ops.zeros([0], dtype='int32'),
                    'feature': keras.ops.zeros([0, 1], dtype='float32')
                }
            ),
            # Graph with three subgraphs where the second subgraph's first node has no edges
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([2, 3, 2], dtype='int32')
                },
                node={
                    'feature': keras.ops.array([[1., 2.], [3., 4.], [5., 6.], [6., 7.], [8., 9.], [10., 11.], [12., 13.]], dtype='float32'),
                    'weight': keras.ops.array([0.50, 1.00, 2.00, 0.25, 0.75, 0.25, 0.25], dtype='float32')
                },
                edge={
                    'source': keras.ops.array([0, 1, 4, 3, 4, 5, 6, 6], dtype='int32'),
                    'target': keras.ops.array([1, 0, 3, 4, 4, 6, 5, 6], dtype='int32'),
                    'feature': keras.ops.array([[1.], [2.], [3.], [4.], [5.], [6.], [7.], [8.]], dtype='float32')
                }
            ),
            # Graph with three subgraphs where the first subgraph's nodes have no edges
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([2, 3, 2], dtype='int32')
                },
                node={
                    'feature': keras.ops.array([[1., 2.], [3., 4.], [5., 6.], [6., 7.], [8., 9.], [10., 11.], [12., 13.]], dtype='float32'),
                    'weight': keras.ops.array([0.50, 1.00, 2.00, 0.25, 0.75, 0.25, 0.25], dtype='float32')
                },
                edge={
                    'source': keras.ops.array([4, 3, 4, 5, 6, 6], dtype='int32'),
                    'target': keras.ops.array([3, 4, 4, 6, 5, 6], dtype='int32'),
                    'feature': keras.ops.array([[3.], [4.], [5.], [6.], [7.], [8.]], dtype='float32')
                }
            ),
            # Graph with three subgraphs where the last subgraph's nodes have no edges
            tensors.GraphTensor(
                context={
                    'size': keras.ops.array([2, 3, 3], dtype='int32')
                },
                node={
                    'feature': keras.ops.array([[1., 2.], [3., 4.], [5., 6.], [6., 7.], [8., 9.], [10., 11.], [12., 13.], [14., 15.]], dtype='float32'),
                    'weight': keras.ops.array([0.50, 1.00, 2.00, 0.25, 0.75, 0.25, 0.25, 0.5], dtype='float32')
                },
                edge={
                    'source': keras.ops.array([0, 1, 4, 3, 4, 5, 6], dtype='int32'),
                    'target': keras.ops.array([1, 0, 3, 4, 4, 6, 5], dtype='int32'),
                    'feature': keras.ops.array([[1.], [2.], [3.], [4.], [5.], [6.], [7.]], dtype='float32')
                }
            ),
        ]

    def test_graph_conv(self):
        for i, tensor in enumerate(self.tensors):
            for drop_edge_feature in [False, True]:
                if drop_edge_feature:
                    tensor = tensor.update(
                        {
                            'edge': {
                                'feature': None
                            }
                        }
                    )
                for skip_connect in [True, False, 'weighted']:
                    for normalize in [True, False, 'batch']:
                        with self.subTest(i=i, skip_connect=skip_connect, normalize=normalize, flat=True):
                            output = layers.GraphConv(128, skip_connect=skip_connect, normalize=normalize)(tensor)
                            self.assertTrue(output.node['feature'].shape[-1] == 128)
                            if not drop_edge_feature:
                                self.assertTrue(output.edge['feature'].shape[-1] == tensor.edge['feature'].shape[-1])

    def test_gi_conv(self):
        for i, tensor in enumerate(self.tensors):
            for drop_edge_feature in [False, True]:
                if drop_edge_feature:
                    tensor = tensor.update(
                        {
                            'edge': {
                                'feature': None
                            }
                        }
                    )
                for skip_connect in [True, False, 'weighted']:
                    for normalize in [True, False, 'batch']:
                        with self.subTest(i=i, skip_connect=skip_connect, normalize=normalize, flat=True):
                            output = layers.GIConv(128, skip_connect=skip_connect, normalize=normalize)(tensor)
                            self.assertTrue(output.node['feature'].shape[-1] == 128)
                            if not drop_edge_feature:
                                self.assertTrue(output.edge['feature'].shape[-1] == tensor.node['feature'].shape[-1])

    def test_ga_conv(self):
        for i, tensor in enumerate(self.tensors):
            for drop_edge_feature in [False, True]:
                if drop_edge_feature:
                    tensor = tensor.update(
                        {
                            'edge': {
                                'feature': None
                            }
                        }
                    )
                for skip_connect in [True, False, 'weighted']:
                    for normalize in [True, False, 'batch']:
                        with self.subTest(i=i, skip_connect=skip_connect, normalize=normalize, flat=True):
                            output = layers.GAConv(128, skip_connect=skip_connect, normalize=normalize)(tensor)
                            self.assertTrue(output.node['feature'].shape[-1] == 128)
                            if not drop_edge_feature:
                                self.assertTrue(output.edge['feature'].shape[-1] == 128)

    def test_mp_conv(self):
        for i, tensor in enumerate(self.tensors):
            for drop_edge_feature in [False, True]:
                if drop_edge_feature:
                    tensor = tensor.update(
                        {
                            'edge': {
                                'feature': None
                            }
                        }
                    )
                for skip_connect in [True, False, 'weighted']:
                    for normalize in [True, False, 'batch']:
                        with self.subTest(i=i, skip_connect=skip_connect, normalize=normalize, flat=True):
                            output = layers.MPConv(128, skip_connect=skip_connect, normalize=normalize)(tensor)
                            self.assertTrue(output.node['feature'].shape[-1] == 128)
                            if not drop_edge_feature:
                                self.assertTrue(output.edge['feature'].shape[-1] == tensor.edge['feature'].shape[-1])

    def test_gt_conv(self):
        for i, tensor in enumerate(self.tensors):
            for drop_edge_feature in [False, True]:
                if drop_edge_feature:
                    tensor = tensor.update(
                        {
                            'edge': {
                                'feature': None
                            }
                        }
                    )
                for skip_connect in [True, False, 'weighted']:
                    for normalize in [True, False, 'batch']:
                        with self.subTest(i=i, skip_connect=skip_connect, normalize=normalize, flat=True):
                            output = layers.GTConv(128, skip_connect=skip_connect, normalize=normalize)(tensor)
                            self.assertTrue(output.node['feature'].shape[-1] == 128)
                            if not drop_edge_feature:
                                self.assertTrue(output.edge['feature'].shape[-1] == tensor.edge['feature'].shape[-1])
                        
    def test_mp_conv3d(self):
        for i, tensor in enumerate(self.tensors):
            tensor = tensor.update(
                {
                    'node': {
                        'coordinate': keras.random.uniform(
                            shape=(tensor.node['feature'].shape[0], 3), minval=-10, maxval=10)
                    }
                }
            )
            for drop_edge_feature in [False, True]:
                if drop_edge_feature:
                    tensor = tensor.update(
                        {
                            'edge': {
                                'feature': None
                            }
                        }
                    )
                for skip_connect in [True, False, 'weighted']:
                    for normalize in [True, False, 'batch']:
                        with self.subTest(i=i, skip_connect=skip_connect, normalize=normalize, flat=True):
                            layer = layers.MPConv(128, skip_connect=skip_connect, normalize=normalize)
                            output = layer(tensor)
                            self.assertTrue(layer.has_node_coordinate)
                            self.assertTrue(output.node['feature'].shape[-1] == 128)
                            self.assertTrue(output.node['coordinate'].shape[-1] == tensor.node['coordinate'].shape[-1])
                            if not drop_edge_feature:
                                self.assertTrue(output.edge['feature'].shape[-1] == tensor.edge['feature'].shape[-1])

    def test_eg_conv3d(self):
        for i, tensor in enumerate(self.tensors):
            tensor = tensor.update(
                {
                    'node': {
                        'coordinate': keras.random.uniform(
                            shape=(tensor.node['feature'].shape[0], 3), minval=-10, maxval=10)
                    }
                }
            )
            for drop_edge_feature in [False, True]:
                if drop_edge_feature:
                    tensor = tensor.update(
                        {
                            'edge': {
                                'feature': None
                            }
                        }
                    )
                for skip_connect in [True, False, 'weighted']:
                    for normalize in [True, False, 'batch']:
                        with self.subTest(i=i, skip_connect=skip_connect, normalize=normalize, flat=True):
                            layer = layers.EGConv(128, skip_connect=skip_connect, normalize=normalize)
                            output = layer(tensor)
                            self.assertTrue(layer.has_node_coordinate)
                            self.assertTrue(output.node['feature'].shape[-1] == 128)
                            self.assertTrue(output.node['coordinate'].shape[-1] == tensor.node['coordinate'].shape[-1])
                            if not drop_edge_feature:
                                self.assertTrue(output.edge['feature'].shape[-1] == tensor.edge['feature'].shape[-1])

    def test_gt_conv3d(self):
        for i, tensor in enumerate(self.tensors):
            tensor = tensor.update(
                {
                    'node': {
                        'coordinate': keras.random.uniform(
                            shape=(tensor.node['feature'].shape[0], 3), minval=-10, maxval=10)
                    }
                }
            )
            for drop_edge_feature in [False, True]:
                if drop_edge_feature:
                    tensor = tensor.update(
                        {
                            'edge': {
                                'feature': None
                            }
                        }
                    )
                for skip_connect in [True, False, 'weighted']:
                    for normalize in [True, False, 'batch']:
                        with self.subTest(i=i, skip_connect=skip_connect, normalize=normalize, flat=True):
                            layer = layers.GTConv(128, skip_connect=skip_connect, normalize=normalize)
                            output = layer(tensor)
                            self.assertTrue(layer.has_node_coordinate)
                            self.assertTrue(output.node['feature'].shape[-1] == 128)
                            self.assertTrue(output.node['coordinate'].shape[-1] == tensor.node['coordinate'].shape[-1])
                            if not drop_edge_feature:
                                self.assertTrue(output.edge['feature'].shape[-1] == tensor.edge['feature'].shape[-1])


if __name__ == '__main__':
    unittest.main()